import sys
from decimal import Decimal
from datetime import datetime
import numpy as np
import os
from io import StringIO
from utils_ak.builtin import delistify
from utils_ak.numeric import is_int, is_float
import yaml


# ultimate json tool for convenient json handling and stuff

# todo: datetime support
# todo: bson.json_util?
# todo: simplejson?

if sys.version_info[:2] == (2, 6):
    # In Python 2.6, json does not include object_pairs_hook. Use simplejson
    # instead.
    try:
        import simplejson as json
    except ImportError:
        import json
else:
    import json


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'to_json'):
            return obj.to_json()
        elif isinstance(obj, Decimal):
            return str(obj)
        elif isinstance(obj, datetime):
            return str(obj)
        elif is_int(obj):
            return int(obj)
        elif is_float(obj):
            return float(obj)
        else:
            return super().default(obj)


def validate(data):
    try:
        return json.loads(data)
    except ValueError:
        pass


def cast_js(js_obj, *args, **kwargs):
    if isinstance(js_obj, str):
        return js_obj
    else:
        return json.dumps(js_obj, cls=JsonEncoder, ensure_ascii=False, *args, **kwargs)


def cast_dict_or_list(js_obj, *args, **kwargs):
    if isinstance(js_obj, (dict, list)):
        return js_obj

    if isinstance(js_obj, bytes):
        try:
            js_obj = js_obj.decode('utf-8')
        except:
            raise Exception('Unknown bytes encoding')

    # load object from file if path exists
    if isinstance(js_obj, str):
        if os.path.exists(js_obj):
            with open(js_obj, 'r') as f:
                js_obj = f.read()

        try:
            res = json.loads(js_obj, *args, **kwargs)
            if isinstance(res, (dict, list)):
                return res
        except:
            pass

        # todo: put away form js file - non-intuitive at all
        # try load as yaml
        try:
            res = delistify(list(yaml.load_all(StringIO(js_obj))))
            if isinstance(res, (dict, list)):
                return res
        except:
            pass

    raise Exception('Unknown type')


dumps = cast_js
loads = cast_dict_or_list




if __name__ == '__main__':
    print(dumps({'a': 'foo'}))
    print(loads(dumps({'a': 'foo', 'dec': Decimal('10.1'), 'today': datetime.now()})))
    print(cast_dict_or_list('columns'))
    print(cast_js(None))

    import apache_beam.transforms.window


import subprocess
import os


def go_build(path, force_rebuild=False):
    filename = os.path.splitext(path)[0]
    rebuild_flag = "-a" if force_rebuild else ""
    command = f"go build {rebuild_flag} -o {filename} {path}"
    print(command)
    subprocess.call(command, shell=True)


def go_run(path, verbatim=True, **kwargs):
    filename = os.path.splitext(path)[0]
    command = f"{filename}"
    for k, v in kwargs.items():
        arg = f" -{k}=\"{v}\""
        command += arg
    print(command)
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    if verbatim:
        while True:
            line = process.stdout.readline().rstrip()
            if not line:
                break
            print(line.decode("utf-8"))


if __name__ == "__main__":
    go_build("/mnt/granular_storage/backoffice.go")
    go_run("/mnt/granular_storage/backoffice.go")


# todo: retry on slack communication
# todo: use cache or not?
# todo: clean cache
# message structure: https://api.slack.com/docs/messages
# todo: make properly
# HARDCODE to implement retry logic inside Slacker reflexion

""" Send and update messages to slack transparently using message ids. """
import json
import os
from datetime import datetime
import requests
from retrypy import retry

import slacker
from slacker import *
from slacker import API_BASE_URL


@retry.decorate(requests.exceptions.RequestException, times=7, wait=lambda n: 2 ** n)
def _request(self, method, api, **kwargs):
    if self.token:
        kwargs.setdefault('params', {})['token'] = self.token

    response = method(API_BASE_URL.format(api=api),
                      timeout=self.timeout,
                      proxies=self.proxies,
                      **kwargs)

    response.raise_for_status()

    response = Response(response.text)
    if not response.successful:
        raise Error(response.error)

    return response


slacker.BaseAPI._request = _request


class Attachment(object):
    def __init__(self, **kwargs):
        self.attachment_text = None
        self.pretext = None
        self.fallback = None
        self.color = None
        self.title = None
        self.title_link = None
        self.author_name = None
        self.author_link = None
        self.author_icon = None
        self.footer = None
        self.footer_icon = None
        self.ts = None
        self.fields = []

        for k, v in kwargs.items():
            setattr(self, k, v)

    def add_field(self, title, value, short=True):
        self.fields.append({'title': title, 'value': value, 'short': short})

    def to_json(self):
        return {
            "fallback": self.fallback,
            "color": self.color,
            "title": self.title,
            "title_link": self.title_link,
            "author_name": self.author_name,
            "author_link": self.author_link,
            "author_icon": self.author_icon,
            "fields": self.fields,
            "footer": self.footer,
            "pretext": self.pretext,
            "text": self.attachment_text,
            'ts': self.ts
        }


class SlackMessage(object):
    def __init__(self, **kwargs):
        self.text = None
        self.attachments = []

        for k, v in kwargs.items():
            setattr(self, k, v)

        if not self.attachments:
            self.attachments = []

        if self.attachments:
            self.attachments = [self._cast_attachment(attachment) for attachment in self.attachments]

    def _cast_attachment(self, attachment):
        if isinstance(attachment, Attachment):
            return attachment
        elif isinstance(attachment, dict):
            return Attachment(**attachment)
        else:
            raise Exception('Unknown attachment type')

    def add_attachment(self, **kwargs):
        self.attachments.append(Attachment(**kwargs))

    def to_json(self):
        res = {
            "text": self.text,
            "attachments": [attachment.to_json() for attachment in self.attachments]
        }

        # clear unused fields
        if not self.text:
            res.pop('text')
        res['attachments'] = [{k: v for k, v in attachment.items() if v is not None} for attachment in res['attachments']]
        res['attachments'] = [attachment for attachment in res['attachments'] if attachment]
        if not res['attachments']:
            res.pop('attachments')

        return res

    def __str__(self):
        return str(self.to_json())

    def __repr__(self):
        return str(self.to_json())


class SimpleMessage(SlackMessage):
    """ Message with one attachment. """

    def __init__(self, **kwargs):
        super().__init__(text=kwargs.pop('text', None), attachments=None)
        self.add_attachment(**kwargs)

    def add_field(self, *args, **kwargs):
        self.attachments[0].add_field(*args, **kwargs)


class SlackMessenger(Slacker):
    def __init__(self, token):
        super().__init__(token)
        self.token = token
        # dict(<message_id>-<channel>: <ts>)
        self.message_timestamps = {}
        self.channel_ids = {}

        self.cache_fn = 'slack_messenger_cache.json'
        self.load_cache()

    def load_cache(self):
        if os.path.exists(self.cache_fn):
            with open(self.cache_fn, 'r') as f:
                self.cache = json.load(f)

            if self.token != self.cache.get('token'):
                # reset cache when slack changes
                self.cache = {}
            else:
                self.message_timestamps = self.cache['message_timestamps']
                self.channel_ids = self.cache['channel_ids']
        else:
            self.cache = {}

    def dump_cache(self):
        self.cache['message_timestamps'] = self.message_timestamps
        self.cache['channel_ids'] = self.channel_ids
        self.cache['token'] = self.token

        with open(self.cache_fn, 'w') as f:
            json.dump(self.cache, f, indent=1)

    def upd_message(self, slack_message, channel, message_id=None):
        # Update message if it exists. Otherwise, post a new message
        """
        :param slack_message: `SlackMessage` or dict or str. In the latter case, str will be used as a text of the message
        :param message_id: key of the message. If the key is the same in 2 consecutive calls, the latter message will be updated. If not specified, the unique message id will be generated
        :param channel: slack channel
        :return:
        """
        if not message_id:
            message_id = str(datetime.now())

        full_id = "{}-{}".format(message_id, channel)

        try:
            # get channel_id
            if channel in self.channel_ids:
                channel_id = self.channel_ids[channel]
            else:
                channel_id = self.get_channel_id_by_name(channel)
                self.channel_ids[channel] = channel_id
        except:
            # channel not found. Possibly, channel is a direct user link (@username)
            channel_id = channel

        # convert slack_message to dict
        if isinstance(slack_message, SlackMessage):
            message_js = slack_message.to_json()
        elif isinstance(slack_message, dict):
            message_js = slack_message
        elif isinstance(slack_message, str):
            message_js = SlackMessage(text=slack_message).to_json()
        else:
            raise Exception('Slack message should be one of: SlackMessage, dict, str')

        message_ts = self.message_timestamps.get(full_id)

        if not message_ts:
            res = self.chat.post_message(text=message_js.get('text'), channel=channel_id, attachments=message_js.get('attachments'))
            self.message_timestamps[full_id] = res.body['ts']
            self.dump_cache()
        else:
            try:
                self.chat.update(text=message_js.get('text'), channel=channel_id, attachments=message_js.get('attachments'), ts=message_ts)
            except Exception as e:
                if str(e) == 'message_not_found':
                    res = self.chat.post_message(text=message_js.get('text'), channel=channel_id, attachments=message_js.get('attachments'))
                    self.message_timestamps[full_id] = res.body['ts']
                    self.dump_cache()
                else:
                    # todo: handle properly
                    print("Unknown error    !", e)

    def get_channels(self):
        return self.channels.list().body['channels']

    def get_channel_id_by_name(self, channel_name):
        if channel_name[0] == '#':
            channel_name = channel_name[1:]
        channels = self.get_channels()
        channel = [ch for ch in channels if ch['name_normalized'] == channel_name][0]
        return channel['id']



import copy

from utils_ak.numeric import *
from utils_ak.simple_vector import *


from utils_ak.block_tree.parallelepiped_block import ParallelepipedBlock
from utils_ak.block_tree.block import Block
from utils_ak.block_tree.validation import validate_disjoint_by_axis
from utils_ak.block_tree.pushers.pushers import simple_push

from loguru import logger

def dummy_push(parent, block, validator, max_tries=24, start_from='last_end', iter_props=None):
    axis = parent.props['axis']

    if is_int_like(start_from):
        cur_start = int(float(start_from))
    elif start_from == 'beg':
        cur_start = block.props['x_rel'][axis]
    else:
        if not parent.children:
            cur_start = 0
        else:
            if start_from == 'last_beg':
                cur_start = max([child.props['x_rel'][axis] for child in parent.children])
            elif start_from == 'last_end':
                cur_start = max([(child.props['x_rel'] + child.size)[axis] for child in parent.children])
            else:
                raise Exception('Unknown beg type')

    iter_props = iter_props or [{}]

    cur_x = cast_simple_vector(block.n_dims)

    cur_try = 0

    # logger.debug('Inserting blocks', parent=parent.props['cls'], block=block.props['cls'])

    while cur_try < max_tries:
        dispositions = []
        for props in iter_props:
            # try to push
            props = copy.deepcopy(props)
            cur_x[axis] = cur_start
            props['x'] = cur_x
            # logger.debug('Trying to push from ', cur_start=cur_start)
            res = simple_push(parent, block, validator=validator, new_props=props)

            if isinstance(res, Block):
                # success
                # logger.debug('Success')
                return block
            else:
                assert isinstance(res, dict)

                if 'disposition' in res:
                    dispositions.append(res['disposition'])

        disposition = min(dispositions) if len(dispositions) == len(iter_props) else 1
        # logger.debug('Disposition', disposition=disposition)
        cur_start += disposition
    raise Exception('Failed to push element')


def test_dummy_push():
    from utils_ak.loguru import configure_loguru_stdout
    configure_loguru_stdout()

    def brute_validator(parent, block):
        for c in parent.children:
            validate_disjoint_by_axis(c, block, axis=parent.props['axis'])

    logger.debug('Dummy push test')
    root = ParallelepipedBlock('root', n_dims=1, x=[2], axis=0)
    a = ParallelepipedBlock('a', n_dims=1, size=[4], axis=0)
    b = ParallelepipedBlock('b', n_dims=1, size=[3], axis=0)
    dummy_push(root, a, brute_validator)
    dummy_push(root, b, brute_validator, start_from=0)
    logger.debug('Root', root=root)


if __name__ == '__main__':
    test_dummy_push()


from .js import dumps, loads
from .serializer import Serializer


class JsonSerializer(Serializer):
    def __init__(self, encoding='utf-8', indent=None):
        self.encoding = encoding
        self.indent = indent

    def encode(self, obj):
        s = dumps(obj, indent=self.indent)
        if self.encoding:
            return s.encode(self.encoding)
        else:
            return s

    def decode(self, s):
        return loads(s)



from utils_ak.fluid_flow import *
import warnings
warnings.filterwarnings('ignore')

def test_flow_processor_1():
    container = Container('Input', value=100, max_pressures=[None, 10])

    processor = Processor('Output')

    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


def test_flow_processor_2():
    container = Container('Input', value=100)
    processor = Processor('Output', processing_time=5, max_pressures=[10, None], transformation_factor=2.)
    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


def test_flow_processor_zero_pressure():
    container = Container('Input', value=100)
    processor = Processor('Output', max_pressures=[0, 0])
    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


def test_flow_processor_limit():
    container = Container('Input', value=100, max_pressures=[None, 10])
    processor = Processor('Output', limits=[50, None])
    pipe_connect(container, processor)
    flow = FluidFlow(container, verbose=True)
    run_flow(flow)


if __name__ == '__main__':
    test_flow_processor_1()
    test_flow_processor_2()
    test_flow_processor_zero_pressure()
    test_flow_processor_limit()


import json
from pathlib import Path
import os


def write_json(path, json_string):
    dir_path = Path(os.path.dirname(path))
    dir_path.mkdir(parents=True, exist_ok=True)
    with open(path, 'w') as outfile:
        json.dump(json_string, outfile)


def read_json(path):
    with open(path, 'r') as infile:
        return json.load(infile)


import fire
from functools import partial
from utils_ak.mongo_job_queue.worker.test.test_worker import TestWorker
from utils_ak.mongo_job_queue.worker.worker import *

if __name__ == '__main__':
    fire.Fire(partial(run_worker, worker_cls=TestWorker))



import json
import os
from utils_ak.granular_storage.table.timeseries.fs_basic import BasicFsTimeSeriesTable


class FsJsonTimeSeriesTable(BasicFsTimeSeriesTable):
    def __init__(self, fn, encoding='utf-8'):
        super().__init__(fn)
        self.fn = fn
        self.encoding = encoding

    def encode(self, msg):
        return (json.dumps(msg) + '\n').encode(encoding=self.encoding)

    def read(self):
        if not os.path.exists(self.fn):
            raise StopIteration

        with open(self.fn, 'rb') as f:
            for line in f:
                yield json.loads(line, encoding=self.encoding)


if __name__ == '__main__':
    from datetime import datetime

    storage = FsJsonTimeSeriesTable('tmp2.tss')

    for i in range(3):
        storage.store(datetime.now().timestamp(), 'key', {'a': i})

    # flush
    storage.close()
    for value in storage.read():
        print(1, value)
    print()

    storage.store_many([[datetime.now().timestamp(), 'key', {'b': i}] for i in range(3)])

    # flush
    storage.close()

    for value in storage.read():
        print(2, value)
    print()
    storage.store_many([[datetime.now().timestamp(), 'key', {'c': i}] for i in range(3)], safe=True)

    # already flushed!

    for value in storage.read():
        print(3, value)
    print()

    storage.clear()



import sys
import math
import json

import stackprinter
from loguru import logger


def serialize(record):
    extra = dict(record['extra'])
    extra.pop('source', None)

    simplified = {
        'level': record['level'].name,
        'message': record['message'],
        'ts': int(math.floor(record['time'].timestamp() * 1000)),  # epoch millis
        'inner_source': record['extra'].get('source', ''),
        'extra': extra,
        'stack': '',
        'error': ''
    }

    exception_info = sys.exc_info()
    if exception_info[0]:
        simplified['stack'] = stackprinter.format(record["exception"])
        simplified['error'] = simplified['stack'].split('\n')[-1]
    return json.dumps(simplified)


def patch_serialized_generic_extra(record):
    record['extra']['serialized'] = serialize(record)


def patch_trace(record):
    exception_info = sys.exc_info()
    if exception_info[0]:
        record['extra']['stack'] = stackprinter.format(record["exception"])
        record['extra']['error'] = record['extra']['stack'].split('\n')[-1]



STUB = '__stub'

class SimpleIterator:
    def __init__(self, lst, start_from=0):
        self.lst = lst
        self.current_index = start_from

    def current(self):
        return self.lst[self.current_index]

    def __len__(self):
        return len(self.lst)

    def forward(self, step=1, return_last_if_out=False, update_index=True, out_value=None):
        if self.current_index >= len(self.lst) - step:
            if return_last_if_out:
                res = self.lst[-1]
            else:
                res = out_value
        else:
            res = self.lst[self.current_index + step]

        if update_index:
            self.current_index = min(self.current_index + step, len(self.lst) - 1)
        return res

    def next(self, return_last_if_out=False, update_index=True):
        return self.forward(1, return_last_if_out, update_index)

    def backward(self, step=1, return_first_if_out=False, update_index=True, out_value=None):
        if self.current_index <= step - 1:
            if return_first_if_out:
                res = self.lst[0]
            else:
                res = out_value
        else:
            res = self.lst[self.current_index - 1]

        if update_index:
            self.current_index = max(self.current_index - step, 0)
        return res

    def prev(self, return_first_if_out=False, update_index=True):
        return self.backward(1, return_first_if_out, update_index)

    def iter(self, direction='up', step=1, limit=None):
        yield self.current()
        counter = 1
        while True:
            if counter == limit:
                break

            if direction == 'up':
                next = self.forward(step, out_value=STUB)
            elif direction == 'down':
                next = self.backward(step, out_value=STUB)

            if next == STUB:
                break

            yield next
            counter += 1



    def iter_sequences(self, n=2):
        for i in range(len(self) - n + 1):
            yield self.lst[i: i + n]

    def reset(self):
        self.current_index = 0


def test_simple_bounded_iterator():
    lst = [1, 2, 3, 4]
    it = SimpleIterator(lst)

    for v in it.iter('up'):
        print(v)
    print()
    for v in it.iter('down'):
        print(v)
    print()
    for i in range(5):
        print(it.next(return_last_if_out=False))
    print()
    for i in range(5):
        print(it.prev(return_first_if_out=True))

    for seq in it.iter_sequences(2):
        print(seq)

    it.reset()

    for v in it.iter(step=2):
        print(v)
    it.reset()

    for v in it.iter(limit=2):
        print(v)


if __name__ == '__main__':
    test_simple_bounded_iterator()



""" Files are simple wrappers to atom file format. """
from .csv import CSVFile
from .feather import FeatherFile
from .hdf import HDFFile
from .msgpack import MsgPackFile
from .stream import StreamFile
from .parquet import ParquetFile



from utils_ak.time.dt import *
from utils_ak.os import list_files, remove_path, make_directories
from utils_ak.granular_storage.df.gran_df.granular import GranularFeather, GranularCSV, GranularMsgPack, GranularParquet
from utils_ak.granular_storage.df.gran_df.granular import LEVEL_ZERO, LEVEL_YEAR, LEVEL_HOUR
from utils_ak.granular_storage.df.gran_df.streamer import GranularStreamer
import numpy as np
import os

# todo: remove mode='w' for safety?


class GranularStorage(object):
    def __init__(self, root='.'):
        self.root = self._fix_root(root)
        make_directories(os.path.dirname(self.root))

    def __getitem__(self, item):
        return self.get_granular(item)

    def granular_names(self):
        fns = list_files(self.root, pattern='*.gran_df', recursive=True)
        names = [os.path.dirname(fn) for fn in fns]
        names = [self._fix_sep(name) for name in names]
        names = [name.replace(self.root, '') for name in names]
        return names

    def exists(self, path):
        return os.path.exists(self._metadata_fn(path))

    def remove(self, path):
        remove_path(self._full_path(path))

    def _full_path(self, path):
        path = self._fix_sep(path)
        if path.startswith(self.root):
            return path
        return os.path.join(self.root, path)

    def get_granular(self, path, level=None, pattern=None, mode='a'):
        return self.cls()(self._full_path(path), level, pattern, mode=mode)

    def cls(self):
        raise NotImplementedError('Not implemented')

    def _metadata_fn(self, path):
        return os.path.join(self._full_path(path), '.gran_df')

    def _fix_sep(self, path):
        path = path.replace('\\', '/')
        path = path.replace(os.path.sep, '/')
        return path

    def _fix_root(self, root):
        if root[-1] != '/':
            root += '/'
        return self._fix_sep(root)

    def get_streamer(self, path, level=3, stream_level=3, **kwargs):
        return GranularStreamer(self, path, level=level, stream_level=stream_level, **kwargs)


class GranularFeatherStorage(GranularStorage):
    def cls(self):
        return GranularFeather


class GranularCSVStorage(GranularStorage):
    def cls(self):
        return GranularCSV


class GranularMsgPackStorage(GranularStorage):
    def cls(self):
        return GranularMsgPack


class GranularParquetStorage(GranularStorage):
    def cls(self):
        return GranularParquet


if __name__ == '__main__':
    gs = GranularCSVStorage('gs_root')

    gran1 = gs.get_granular('gran1', level=LEVEL_ZERO)
    gran2 = gs.get_granular('gran2', level=LEVEL_YEAR)
    gran3 = gs.get_granular('gran3', level=LEVEL_HOUR)

    index = [cast_datetime('2018.05.01') + timedelta(days=1) * i for i in range(3)]
    df1 = pd.DataFrame(np.random.randn(3, 1), index=index, columns=['0'])

    gran1.append(df1, 'foo1')
    gran2.append(df1, 'foo2')
    gran3.append(df1, 'foo3')

    print(gs.granular_names())

    for name in gs.granular_names():
        gran = gs[name]
        for key in gran.keys():
            print(key)
            print(gran[key].head(3))

    gs.remove('gran3')
    print(gs.granular_names())



import time
import asyncio
import inspect

from dateutil import rrule
from datetime import datetime, timedelta
from utils_ak.time import cast_sec, cast_timedelta

TIME_EPS = 0.001


class CallbackTimer:
    """ Run callback function every interval. """

    def __init__(self, callback, interval, timer_type="right", args=None, kwargs=None):
        """
        :param timer_type: str, "left" to start timing immediately, "right" to wait for callback() to finish
        """
        self.callback = callback
        self.interval = cast_sec(interval)
        self.last_call = 0
        self.timer_type = timer_type

        self.args = args or tuple()
        self.kwargs = kwargs or {}

        if timer_type not in ['right', 'left']:
            raise Exception('Only right/left timer is supported')

    @property
    def next_call(self):
        return self.last_call + self.interval

    def run_if_possible(self):
        if time.time() > self.next_call:
            if self.timer_type == "right":
                try:
                    self.run()
                except:
                    self.last_call = time.time()
                    raise
                else:
                    self.last_call = time.time()

            elif self.timer_type == "left":
                self.last_call = time.time()
                self.run()
            return True
        return False

    def run(self):
        return self.callback(*self.args, **self.kwargs)

    async def run_if_possible_aio(self):
        if time.time() > self.next_call:
            if self.timer_type == "right":
                try:
                    await self.run_aio()
                except:
                    self.last_call = time.time()
                    raise
                else:
                    self.last_call = time.time()
            elif self.timer_type == "left":
                self.last_call = time.time()
                await self.run_aio()
            return True
        return False

    async def run_aio(self):
        if inspect.iscoroutinefunction(self.callback):
            return await self.callback(*self.args, **self.kwargs)
        else:
            return self.callback(*self.args, **self.kwargs)


class ScheduleTimer:
    """ Run callback function at schedule. """
    def __init__(self, callback, pattern=None, freq=None, init_run=False, args=None, kwargs=None):
        """
        :param pattern: (seconds) (minutes) (hours) (day) (month) (day-of-week) in cron-like manner: * 5 * * * * ...
        """
        self.callback = callback
        self.next_call = 0
        self.init_run = init_run

        self.args = args or tuple()
        self.kwargs = kwargs or {}

        if freq:
            # cast pattern by frequency
            if cast_timedelta(freq) >= timedelta(days=1):
                # freq is supported only for hour and less frequency
                raise Exception(f'Too large frequency: {cast_sec(freq)}. Use frequency less than a day')
            freq_sec = round(cast_sec(freq))

            if freq_sec <= 0:
                raise Exception('Bad frequency: {}'.format(freq_sec))

            unit_secs = [1, 60, 3600, 3600 * 24]

            # find which unit is used
            unit_i = 0
            while True:
                if unit_secs[unit_i] <= freq_sec < unit_secs[unit_i + 1]:
                    break
                unit_i += 1

            if unit_secs[unit_i + 1] % freq_sec != 0:
                raise Exception('Bad frequency: {}'.format(freq_sec))

            freq_units = int(freq_sec / unit_secs[unit_i])
            freq_vals = [i * freq_units for i in range(unit_secs[unit_i + 1] // freq_sec)]
            freq_vals = map(str, freq_vals)

            pattern_vals = []
            for i in range(6):
                if i < unit_i:
                    pattern_vals.append('0')
                elif i == unit_i:
                    if freq_units != 1:
                        pattern_vals.append(','.join(freq_vals))
                    else:
                        pattern_vals.append('*')
                elif i > unit_i:
                    pattern_vals.append('*')
            self.pattern = ' '.join(pattern_vals)
        else:
            self.pattern = pattern
        units = sec, min, hou, day, mon, dow = [self._parse_time_unit(unit) for unit in self.pattern.split()]
        freqs = [rrule.SECONDLY, rrule.MINUTELY, rrule.HOURLY, rrule.DAILY, rrule.MONTHLY, rrule.DAILY]
        param_names = ['bysecond', 'byminute', 'byhour', 'bymonthday', 'bymonth', 'byweekday']

        try:
            freq = freqs[units.index(None)]
        except:
            # not stars at all
            freq = rrule.YEARLY

        self.rrule_params = dict(zip(param_names, units))
        self.rrule_params['freq'] = freq

        self._upd_next_call()

    def _parse_time_unit(self, unit):
        if unit == '*':
            return None
        unit = unit.split(',')
        return [int(x) for x in unit]

    def _upd_next_call(self):
        # set next_call to closest
        self.last_call_dt = datetime.fromtimestamp(self.next_call) if self.next_call else datetime.now()
        next_call_lst = list(rrule.rrule(dtstart=self.last_call_dt + timedelta(seconds=1),
                                         until=self.last_call_dt + timedelta(seconds=1) + timedelta(days=366),
                                         count=1,
                                         **self.rrule_params))

        if not next_call_lst:
            raise Exception('Bad scheduler input!')

        self.next_call = next_call_lst[0].timestamp()

    def check(self):
        if self.init_run or time.time() > self.next_call:
            self.init_run = False
            self.run()
            self._upd_next_call()
            return True
        return False

    async def aiocheck(self):
        if self.init_run or time.time() > self.next_call:
            self.init_run = False
            await self.aiorun()
            self._upd_next_call()

    def run(self):
        return self.callback(*self.args, **self.kwargs)

    async def aiorun(self):
        if inspect.iscoroutinefunction(self.callback):
            return await self.callback(*self.args, **self.kwargs)
        else:
            return self.callback(*self.args, **self.kwargs)


class CallbackTimers(object):
    """ A builtin of callback timers. """

    def __init__(self):
        self.timers = []

    def add_timer(self, timer):
        self.timers.append(timer)

    def check(self):
        if time.time() > self.next_call:
            for timer in self.timers:
                timer.run_if_possible()

    @property
    def next_call(self):
        return min([timer.next_call for timer in self.timers])

    def next_call_timeout(self):
        return max(self.next_call - time.time(), 0.)

    def __len__(self):
        return len(self.timers)

    async def aiocheck(self):
        if time.time() > self.next_call:
            for timer in self.timers:
                await timer.run_if_possible()


def ex_sequential():
    def print_msg(msg=None):
        print('Callback:', datetime.now(), msg)

    print('Timer 1')
    timer1 = CallbackTimer(print_msg, 1)
    timer2 = CallbackTimer(print_msg, 1, args=('asdf',))
    for i in range(300):
        timer1.run_if_possible()
        timer2.run_if_possible()
        time.sleep(0.01)

    print('Timer 2')
    # run every second
    timer1 = ScheduleTimer(print_msg, '* * * * * *')
    timer2 = ScheduleTimer(print_msg, '* * * * * *', args=('asdf',))
    for i in range(300):
        timer1.check()
        timer2.check()
        time.sleep(0.01)

    print('Timer 3')
    # run every second
    timer = ScheduleTimer(print_msg, freq='2s', init_run=True)
    print(timer.pattern)
    N = 3 * 1000
    for i in range(N):
        timer.check()
        time.sleep(1 / 10000)


def ex_coroutine():
    import numpy as np
    async def print_msg(msg=None):
        timeout = np.random.uniform(0, 1)
        await asyncio.sleep(timeout)
        print('Callback:', datetime.now(), msg, timeout)

        if timeout > 0.5:
            raise Exception('Test')

    timer = CallbackTimer(print_msg, 1.0)

    async def wrapper():
        while True:
            try:
                await timer.coroutine()
            except Exception as e:
                print(e)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.ensure_future(wrapper()))
    loop.close()


if __name__ == '__main__':
    ex_coroutine()



import pandas as pd
import os

from utils_ak.split_file import SplitFile


class PandasCSVWriter:
    def __init__(self, fn, buffer_length=1e5, max_file_size=1e9):
        self.buffer_length = buffer_length
        self.fn = fn
        self.buffer_values = []
        self.max_file_size = max_file_size
        self.header = None

    def write_header(self, header):
        self.header = header

    def write_values(self, values):
        self.buffer_values.extend(values)

        if self.buffer_length is not None and len(values) >= self.buffer_length:
            self.flush()

    def flush(self):
        sf = SplitFile(self.fn)
        fn = sf.get_current(max_size=self.max_file_size)
        if os.path.exists(fn):
            pd.DataFrame(self.buffer_values, columns=self.header).to_csv(fn, mode='a', index=False, header=False)
        else:
            pd.DataFrame(self.buffer_values, columns=self.header).to_csv(fn, mode='a', index=False)
        self.reset_buffer()

    def reset_buffer(self):
        self.buffer_values = []



import os
import anyconfig
import collections
import yaml
import sys

from utils_ak.dict import dotdict
from utils_ak.builtin import update_dic


def cast_config(obj, required=False):
    if obj is None:
        return {}
    elif isinstance(obj, collections.abc.Mapping):
        return obj
    elif isinstance(obj, str):
        if os.path.exists(obj):
            return anyconfig.load(obj)
        else:
            if required:
                raise Exception('Config file not found {}'.format(obj))
            else:
                return {}
    elif isinstance(obj, list):
        res = {}
        for v in obj:
            anyconfig.merge(res, cast_config(v), ac_merge=anyconfig.MS_DICTS)
        return res
    else:
        raise Exception('Unknown config type')


def get_config(configs=None, require_local=False, global_configs=('common_config.yml', 'secret_config.yml', 'instance_config.yml')):
    local_config_fn = os.path.splitext(os.path.abspath(sys.argv[0]))[0] + '.yml'

    if not os.path.exists(local_config_fn):
        if require_local:
            raise Exception(f'Local config not found {local_config_fn}')
        local_config_fn = None

    res = []
    # add global configs
    for base_config in global_configs:
        for dirname in [os.getcwd(), os.path.dirname(os.path.abspath(sys.argv[0]))]:
            fn = os.path.join(dirname, base_config)
            if os.path.exists(fn):
                res.append(fn)
                break

    configs = configs or []
    res += configs

    # add local config
    res.append(local_config_fn)

    return cast_config(res)


config = get_config()


if __name__ == '__main__':
    print(config)


import telegram
import enum


# todo: refactor


class EventType(enum.Enum):
    disconnected = 'DISCONNECTED'
    connection_lost = 'CONNECTION LOST'
    connection_failed = 'CONNECTION FAILED'
    reconnected = "RECONNECTED"
    connected = "CONNECTED"
    snapshot_failed = "SNAPSHOT FAILED"
    gap = "GAP"


class NotifierBot():
    def __init__(self):
        # @qtb_bot token
        self.bot_token = "442059717:AAFUUCQ0mDcoOBYqTZrcoIa8JUY35o82b0E"

        # id of QTB private channel
        self.chat_id = "-1001364676826"

        self.bot = telegram.Bot(token=self.bot_token)

    def notify(self, exchange=None, time=None, errtype=None, curr=None, mss=None):
        if not mss:
            if errtype == EventType.disconnected:
                msg = "{} {}: a disconnect occured on {} at {}. \n Attempting reconnect..." \
                    .format(exchange.upper(), errtype.value, exchange, time)
            elif errtype == EventType.connected:
                msg = "{} {}: successfully connected to {} at {}." \
                    .format(exchange.upper(), errtype.value, exchange, time)
            elif errtype == EventType.reconnected:
                msg = "{} {}: successfully reconnected to {} at {}." \
                    .format(exchange.upper(), errtype.value, exchange, time)
            elif errtype == EventType.gap:
                msg = "{} {}: a gap of more than 1 minute has been observed when downloading from {} at {}".format(exchange.upper(), errtype.value, exchange, time)
            elif errtype == EventType.connection_lost:
                msg = "{} {}: connection lost with {} at {}." \
                    .format(exchange.upper(), errtype.value, exchange, time)
            elif errtype == EventType.connection_failed:
                msg = "{} {}: connection failed with {} at {}." \
                    .format(exchange.upper(), errtype.value, exchange, time)
            elif errtype == EventType.snapshot_failed:
                msg = "{} {}: {} snapshot failed on {} at {}." \
                    .format(exchange.upper(), errtype.value, curr, exchange, time)
            else:
                msg = "{}: UNKNOWN ERROR TYPE ENCOUNTRED ON {} AT {}!".format(exchange.upper(), exchange, time)
        else:
            msg = mss

        self.bot.send_message(self.chat_id, msg)


# Test
if __name__ == "__main__":
    bt = NotifierBot()
    bt.notify("Binance", "time", EventType.reconnect)



import a
import os


class S3Client(object):
    def __init__(self, api_key=None, api_secret=None):
        self.api_key = api_key or None
        self.api_secret = api_secret or None
        self.client = boto3.client('s3',
                                   aws_access_key_id=self.api_key,
                                   aws_secret_access_key=self.api_secret)

    def _file_exists_on_s3(self, file_path, bucket):
        from botocore.errorfactory import ClientError
        """
        Returns True if a file at 'file_path' exists in 'bucket'
        :param file_path: path to target file
        :param bucket: bucket os S3
        """
        try:
            self.client.head_object(Bucket=bucket, Key=file_path)
        except ClientError:
            return False
        except Exception as e:
            raise (e)
        else:
            return True

    def download(self, path, bucket, local=None, overwrite=False, remove=False):
        """
        Downloads a file or a folder from a specified bucket on S3.
        Preserves hierarchy.
        :param path: path to file/folder WITHIN 'bucket'
        :param local: local folder to copy to
        :param bucket: bucket on S3
        :param overwrite: overwrites local files if True
        :param remove: if True, removes the origin file
        """
        local = local or os.getcwd()
        path = path.strip(os.sep)

        paginator = self.client.get_paginator('list_objects')
        for p in paginator.paginate(Bucket=bucket, Delimiter=os.sep, Prefix=path):
            if p.get('CommonPrefixes') is not None:
                for subdir in p.get('CommonPrefixes'):
                    self.download(subdir.get('Prefix') + os.sep, local, bucket)
            if p.get('Contents') is not None:
                for file in p.get('Contents'):
                    fk = file.get('Key')
                    local_path = os.path.join(local, fk)
                    if not os.path.exists(os.path.dirname(local_path)):
                        os.makedirs(os.path.dirname(local_path))
                    if not os.path.exists(local_path) or overwrite:
                        self.client.download_file(bucket, fk, local_path)
                    if remove:
                        self.client.delete_object(Bucket=bucket, Key=fk)

    def upload(self, path, bucket, root=None, target='default', overwrite=False, remove=False):
        """
        Uploads a file or a folder to a specified target location on a bucket on S3.
        Preserves hierarchy.
        :param path: path to a local file/folder WITHIN "top"
        :param root: path to folder where the hierarchy should be copied from (e.g. ../mnt/granular_storage)
        :param target: target folder withing the 'bucket' to copy to
        :param bucket: bucket on S3
        :param overwrite: overwrites files if True
        :param remove: if True, removes the origin file
        """
        # path = path.strip(os.sep)
        len_top = 0
        try:
            target = target.strip(os.sep) + os.sep
            root = root or os.sep + root.strip(os.sep) + os.sep
            path = root + path
            len_top += len(root)
        except:
            pass

        if os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for file in files:
                    local_path = os.path.join(root, file)
                    local_path_rel = local_path[len_top:]
                    file_exists = self._file_exists_on_s3(target + local_path_rel, bucket)
                    if file_exists:
                        if overwrite:
                            self.client.upload_file(local_path, bucket, target + local_path_rel)
                    else:
                        self.client.upload_file(local_path, bucket, target + local_path_rel)
                    if remove:
                        os.remove(os.path.join(root, file))  # Remove files to empty folders
            if remove:  # Remove folders
                for root, dirs, files in os.walk(path):
                    for dir in dirs:
                        os.rmdir(os.path.join(root, dir))
                os.rmdir(root)

        elif os.path.isfile(path):
            file = path[path.rfind(os.sep) + 1:]
            if self._file_exists_on_s3(target + file, bucket):
                if overwrite:
                    self.client.upload_file(path, bucket, target + file)
            else:
                self.client.upload_file(path, bucket, target + file)
            if remove:
                os.remove(path)
        else:
            raise Exception(f'Unknown path: {path}')

    def delete(self, target, bucket):
        """
        :param bucket: bucket name on S3
        :param target: target directory/file to delete
        """
        target = target.strip(os.sep)

        paginator = self.client.get_paginator('list_objects')
        for p in paginator.paginate(Bucket=bucket, Delimiter=os.sep, Prefix=target):
            if p.get('CommonPrefixes') is not None:
                for subdir in p.get('CommonPrefixes'):
                    self.delete(target=subdir.get('Prefix') + os.sep, bucket=bucket)
            if p.get('Contents') is not None:
                delete_dict = dict(Objects=[])
                for file in p.get('Contents'):
                    fk = file.get('Key')
                    delete_dict['Objects'].append(dict(Key=fk))
                    if len(delete_dict['Objects']) > 100:  # There is a limit on chunk size in delete_objects
                        self.client.delete_objects(Bucket=bucket, Delete=delete_dict)
                        delete_dict = dict(Objects=[])
                if len(delete_dict['Objects']) > 0:
                    self.client.delete_objects(Bucket=bucket, Delete=delete_dict)

    def get_all_s3_keys(self, bucket, prefix=None, without_prefix=False):
        """
            Get a list of all keys in an S3 bucket.
            :param bucket: bucket name on S3
            :param prefix: prefix path
        """
        kwargs = {'Bucket': bucket}
        if prefix is not None:
            kwargs['Prefix'] = prefix

        while True:
            resp = self.client.list_objects_v2(**kwargs)
            if 'Contents' in resp:
                for obj in resp['Contents']:
                    key = obj['Key']
                    if without_prefix and os.path.normpath(key) == os.path.normpath(prefix):
                        continue
                    yield key

            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break



from functools import wraps


def add_map(f):
    @wraps(f)
    def wrapper(value, *args, **kwds):
        if isinstance(value, list):
            return [f(x) for x in value]
        else:
            return f(value, *args, **kwds)
    return wrapper


def add_uni(f):
    @wraps(f)
    def wrapper(value, *args, **kwds):
        if isinstance(value, list):
            return f(value, *args, **kwds)
        else:
            return f([value], *args, **kwds)[0]
    return wrapper


if __name__ == '__main__':
    def f(v):
        return v ** 2

    print(add_map(f)(1))
    print(add_map(f)([1, 2, 3]))

    def g(l):
        return [x ** 2 for x in l]

    print(add_uni(g)(1))
    print(add_uni(g)([1, 2, 3]))



import time
from utils_ak.messenger.slack.slack_messenger import SlackMessenger, SimpleMessage, SlackMessage, Attachment

token = 'xoxb-316677261669-rsCabyKYTxii2AlbNgE6T9Yq'
messenger = SlackMessenger(token)


def send_message_example():
    # send with SlackMessage
    message = SimpleMessage(text='Message send test')
    message_dic = message.to_json()
    messenger.upd_message(message, '#general')
    time.sleep(1)
    # send with dict
    messenger.upd_message(message_dic, '#general')
    time.sleep(1)
    # send with str (text)
    messenger.upd_message("Message send test", '#general')


def send_and_update_example():
    messenger.upd_message("Message update test before the update", '#general', "message_unique_key")
    time.sleep(5)
    messenger.upd_message("Message update test after the update", '#general', 'message_unique_key')


def send_attachment_message():
    message = SimpleMessage(text='Message text', ts=time.time() - 300)
    message.add_field("Title 1", "Value 1")
    message.add_field("Title 2", "Value 2")
    message.add_field("Title 3", "Value 3")
    print(message.to_json())
    messenger.upd_message(message, '#general')


def send_general_slack_message_with_attachments():
    message = SlackMessage(text="This is some text")
    attachment1 = Attachment(color='#46B29D', attachment_text='Attachent1')
    attachment1.add_field('Field title', 'Field value')
    attachment2 = Attachment(footer='footer here')
    message.attachments += [attachment1, attachment2]
    messenger.upd_message(message, '#random')


if __name__ == '__main__':
    #     send_message_example()
    #     send_and_update_example()
    #
    #     """ Cache after the run
    #     {
    #  "message_timestamps": {
    #   "2018-02-16 15:31:50.736173-#general": "1518784311.000093",
    #   "2018-02-16 15:31:52.641501-#general": "1518784312.000195",
    #   "2018-02-16 15:31:53.913882-#general": "1518784313.000284",
    #   "2018-02-16 15:32:04.405777-#general": "1518784324.000237",
    #   "2018-02-16 15:32:06.023518-#general": "1518784325.000324",
    #   "2018-02-16 15:32:07.285910-#general": "1518784327.000207",
    #   "message_unique_key-#general": "1518784328.000029"
    #  },
    #  "channel_ids": {
    #   "#general": "C0NKNEZ89"
    #  }
    # }"""
    #
    #     send_attachment_message()
    #     send_general_slack_message_with_attachments()
    message = SimpleMessage(text='hello world')
    messenger.upd_message(message, channel='#hub')
    messenger.upd_message(message, channel='#log')



import logging
import random
import time

from utils_ak.interactive_imports import *

configure_stream_logging(stream_level=logging.INFO)


class Heartbeater(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__(f'Publisher {random.randint(0, 10 ** 6)}', *args, **kwargs)
        self.add_timer(self.publish_json, 1.0, args=('monitor', 'asdf', {'id': self.id},))


if __name__ == '__main__':
    run_listener_async('monitor', message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))
    Heartbeater(message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}})).run()



from .builtin import *
from .callback_timer import *
from .clock import *
from .dict import *
from .exceptions import *
from .functions import *
from .json import *
from .log import *
from .messenger import *
from .numeric import *
from .os import *
from .pandas import *
from .re import *
from .reflexion import *
from .serialization import *
from .str import *
from .time import *
from .tqdm import *
from .jupyter import *
from .simple_microservice import *
from .config import *
from .streaming import *
from .properties import *
from .pandas_writer import *
from .split_file import *



import pandas as pd
import numpy as np

import os
import base64

from PIL import Image
from io import BytesIO


pd.set_option('display.max_colwidth', None)


def display_with_image_thumbnails(df, shape=None):
    from IPython.display import HTML, display

    def _get_thumbnail(path, shape=None):
        if shape is None:
            shape = (150, 150)
        i = Image.open(path)
        i.thumbnail(shape, Image.LANCZOS)
        return i

    def _image_base64(im, shape):
        if isinstance(im, str):
            im = _get_thumbnail(im, shape)
        with BytesIO() as buffer:
            im.save(buffer, 'jpeg')
            return base64.b64encode(buffer.getvalue()).decode()

    def _image_formatter(im, shape):
        if type(im) != str:
            return str(im)
        possible_path = im
        if not os.path.exists(possible_path):
            return im

        ext = os.path.splitext(possible_path)[-1]

        if ext not in ['.png', '.jpg']:
            return im

        return '<img src="granular_storage:image/jpeg;base64,{}">'.format(_image_base64(im, shape))

    columns = df.columns
    formatters = {col: lambda x: _image_formatter(x, shape) for col in columns}
    # displaying PIL.Image objects embedded in dataframe
    display(HTML(df.to_html(formatters=formatters, escape=False)))


def where(cond, x, y):
    """ np.where analogue for dataframes with the same signature. Return elements, either from x or y, depending on condition.

    Parameters
    ----------
    cond: boolean condition
    cond_value: value to return when cond is True
    other: value to return when cond is False

    Returns
    -------
    pd.DataFrame
        np.where wrapped pandas dataframe
    """
    if isinstance(x, pd.DataFrame):
        df = x
    elif isinstance(y, pd.DataFrame):
        df = y
    elif isinstance(cond, pd.DataFrame):
        df = cond
    else:
        raise Exception('Either cond, x, y must be a dataframe')
    return pd.DataFrame(np.where(cond, x, y), index=df.index, columns=df.columns)


def series_equal(s1, s2, dtype='float', eps=1e-8):
    if dtype == 'float':
        return s1.sub(s2).abs() < eps
    else:
        return s1 == s2


def find_row(df, row):
    """ Find last occurence of row in df
    :df: pd.DataFrame
    :row: pd.DataFrame with unit length
    :return: int. Row number
    """
    if len(row) != 1:
        raise Exception('Row should be one line dataframe')

    if not np.all(df.columns == row.columns):
        raise Exception('Dataframe and row have different columns')

    # filter by index
    # time is considered discrete - no problem on rounding here
    df.index.name = 'index'
    df = df.reset_index()
    tmp = df[df['index'] == row.index[0]].copy_path()
    tmp.pop('index')

    if len(tmp) == 0:
        return None

    # filter by values
    for col in tmp.columns:
        # todo: make better
        dtype = 'float' if 'float' in str(df.dtypes[0]) else None

        row_series = pd.Series(index=tmp[col].index)
        row_series[:] = row[col].iloc[0]

        tmp[col] = series_equal(tmp[col], row_series, dtype=dtype)
    ''' 0    False
        1     True
        2    False
        3    False
        dtype: bool'''
    row_equal = tmp.all(axis=1)
    tmp = tmp[row_equal]

    if len(tmp) == 0:
        return None

    return tmp.index[-1]


def merge_by_columns(dfs):
    res_df = dfs[0].copy_path()
    for df in dfs[1:]:
        for key in df.columns:
            res_df[key] = df[key]
    return res_df


def merge(dfs, by=None, by_index=True, keep='last', sort_index=True):
    """
    :param dfs: list(`pd.DataFrame`)
    :param by: name of column or list of columns names. 'all' for all columns. 'columns' for merge_by_columns method
    :param by_index: include index or not (bool value)
    :param keep: 'last' or 'first'

    :return:
    """
    if not by and not by_index:
        raise TypeError('Either by or index should by non-trivial')

    if by is None:
        by = []
    elif by == 'all':
        pass
    elif by == 'columns':
        return merge_by_columns(dfs)
    elif isinstance(by, str):
        by = [by]
    elif isinstance(by, list):
        pass
    else:
        raise TypeError('By should be list, str or None')

    df = pd.concat(dfs, axis=0)

    masks = []
    if by == 'all':
        masks.append(df)
    elif by:
        masks.append(df[by])

    if by_index:
        masks.append(pd.Series(df.index, index=df.index))
    mask = pd.concat(masks, axis=1)
    df = df[~mask.duplicated(keep=keep)]

    if sort_index:
        df = df.sort_index()
    return df


def find_gaps(index, gap):
    index = pd.Series(index, index=index)
    index_diff = index - index.shift(1)
    index_diff = index_diff[index_diff > gap]
    return index_diff


# todo: search for existing solutions
def pd_read(fn, **kwargs):
    ext = os.path.splitext(fn)[-1]
    return getattr(pd, f'read_{ext[1:]}')(fn, **kwargs)


def pd_write(df, fn, **kwargs):
    ext = os.path.splitext(fn)[-1]
    return getattr(df, f'to_{ext[1:]}')(fn, **kwargs)


def mark_consecutive_groups(df, key, groups_key):
    cur_value = None
    cur_i = 0

    values = []
    for i, row in df.iterrows():
        if row[key] != cur_value:
            cur_i += 1
            cur_value = row[key]
        values.append(cur_i)
    df[groups_key] = values

if __name__ == '__main__':
    df1 = pd.DataFrame.from_dict({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [1, 1, 1]})
    df1 = df1.set_index('c')
    df2 = pd.DataFrame.from_dict({'a': [3, 4, 5], 'b': [7, 8, 9]})

    print(df1)
    print('-----')
    print(df2)

    print('----------')

    dfs = [df1, df2]

    print(merge(dfs, index=False, by='a'))
    print()
    print(merge(dfs, index=True, by='a'))
    print()
    print(merge(dfs, index=True, by=None))
    print()
    print(merge(dfs, index=True, by=None, keep='first'))
    print()
    print(merge(dfs, index=True, by='all', keep='first'))

    try:
        print(merge(dfs, index=False, by=None))
    except Exception as e:
        print(e)
    else:
        raise Exception('Should not happen')




""" Python log functionality. """
from .configure import configure_stream_logging, configure_logging, reset_logging
from .simple_custom_field_formatter import SimpleCustomFieldFormatter
from .extra_fields_logger import ExtraFieldsLogger


from utils_ak.granular_storage.df.file.stream import StreamFile
from utils_ak.time.dt import *

import pandas as pd
import os


# todo: debug
# todo: add extensibility

class CSVFile(StreamFile):
    def __init__(self, fn, buffer_size=None, buffer_timeout=60, compression=None, extensible=True):
        self.fn = fn
        self.compression = compression
        self.extensible = extensible
        super().__init__(buffer_size, buffer_timeout)

    def _preprocess(self, df):
        # todo: convert datetime stuff
        df = df.reset_index()
        df['index'] = df['index'].apply(lambda dt: int(cast_mts(dt)))
        df.columns = [str(col) for col in df.columns]
        return df

    def append(self, df, append=True):
        if self.extensible:
            if append and os.path.exists(self.fn):
                df.columns = [str(col) for col in df.columns]
                old_df = self.read()
                df = pd.concat([old_df, df], axis=0)
            df = self._preprocess(df)
            df.to_csv(self.fn, mode='w', compression=self.compression, index=False, header=True)
        else:
            mode = 'a' if append else 'w'
            write_header = mode == 'w' or not os.path.exists(self.fn)
            df = self._preprocess(df)
            df.to_csv(self.fn, mode=mode, compression=self.compression, index=False, header=write_header)

    def append_stream(self, df):
        super().append_stream(df, 'default')

    def read(self, index_col='index'):
        df = pd.read_csv(self.fn)
        df[index_col] = cast_datetime_series(df[index_col])
        df.set_index([index_col], inplace=True)
        df.index.name = 'index'
        df.columns = [str(col) for col in df.columns]
        return df

    def ext(self):
        return '.csv.gz'



class FsTimeSeriesTable:
    def __init__(self):
        pass

    def store(self, ts, key, msg):
        pass

    def store_many(self, values):
        for ts, key, msg in values:
            self.store(ts, key, msg)

    def clear(self):
        pass



class DAGNode:
    def __init__(self):
        self.parents = []
        self.children = []
        self.reset_iteration_state()

    def reset_iteration_state(self, orient='down', with_children=False):
        self.iteration_state = {'is_processed': False}  # iteration state variable

        if with_children:
            for child in self.oriented_children(orient):
                child.reset_iteration_state(orient, with_children=True)

    def leaves(self, orient='down'):
        res = []
        for node in self.iterate_as_tree(orient):
            if not node.oriented_children(orient):
                res.append(node)
        return res

    def top(self, orient='down'):
        leaves = self.leaves(orient)
        assert len(leaves) == 1, 'Top not found'
        if len(leaves) == 1:
            return leaves[0]

    def root(self, orient='down'):
        cur_node = self

        while True:
            if not cur_node.oriented_children(orient):
                return cur_node
            else:
                cur_node = cur_node.oriented_children(orient)[0]

    def oriented_parents(self, orient='down'):
        return {'down': self.parents, 'up': self.children}[orient]

    def oriented_children(self, orient='upwards'):
        return {'down': self.children, 'up': self.parents}[orient]

    def iterate(self, orient='down'):
        assert not self.oriented_parents(orient), 'Can iterate only from root node'
        self.reset_iteration_state(orient, with_children=True)

        cur_node = self

        while True:
            unprocessed_parents = [node for node in cur_node.oriented_parents(orient) if not node.iteration_state['is_processed']]
            if unprocessed_parents:
                cur_node = unprocessed_parents[0]
                continue

            yield cur_node
            cur_node.iteration_state['is_processed'] = True

            unprocessed_children = [node for node in cur_node.oriented_children(orient) if not node.iteration_state['is_processed']]

            if unprocessed_children:
                cur_node = unprocessed_children[0]
                continue
            else:
                break

    @staticmethod
    def _iter_node_as_tree_recursive(cur_node, orient='down'):
        if cur_node.iteration_state['is_processed']:
            return
        else:
            yield cur_node
            cur_node.iteration_state['is_processed'] = True

        for child in cur_node.oriented_children(orient):
            for node in DAGNode._iter_node_as_tree_recursive(child, orient):
                yield node

    def iterate_as_tree(self, orient='down'):
        assert not self.oriented_parents(orient), 'Can iterate only from root node'
        self.reset_iteration_state(orient, with_children=True)

        for node in self._iter_node_as_tree_recursive(self):
            yield node


def connect(parent, child, safe=True):
    if safe and (not parent or not child):
        return
    parent.children.append(child)
    child.parents.append(parent)


def disconnect(parent, child, safe=True):
    if safe and (not parent or not child):
        return
    parent.children.remove(child)
    child.parents.remove(parent)


def test_dag_iteration():
    class NamedNode(DAGNode):
        def __init__(self, name):
            super().__init__()
            self.name = name

        def __str__(self):
            return self.name

        def __repr__(self):
            return self.name

    root_up = NamedNode('root_up')
    node1 = NamedNode('1')
    node2 = NamedNode('2')
    root_down = NamedNode('root_down')

    connect(root_up, node1)
    connect(root_up, node2)
    connect(node1, root_down)
    connect(node2, root_down)

    print('Testing iteration')
    for orient in ['down', 'up']:
        print('Processing orientation', orient)
        for node in [root_up, node1, node2, root_down]:
            print('Processing node', node)
            if orient == 'down' and node == root_up:
                for iter_node in node.iterate(orient):
                    print(iter_node)
            elif orient == 'up' and node == root_down:
                for iter_node in node.iterate(orient):
                    print(iter_node)
            else:
                try:
                    list(node.iterate(orient))
                except AssertionError as e:
                    print('AssertionError', e)

    print('Testing root')
    for orient in ['down', 'up']:
        print('Processing orientation', orient)
        for node in [root_up, node1, node2, root_down]:
            print(node.root(orient))

    print('Testing Leaves, top, iterate as tree')
    root_up = NamedNode('root_up')
    node1 = NamedNode('1')
    node2 = NamedNode('2')

    connect(root_up, node1)
    connect(root_up, node2)
    print(list(root_up.leaves()))
    try:
        print(root_up.top())
    except AssertionError as e:
        print('AssertionError', e)
    top = NamedNode('Top')
    for leaf in root_up.leaves():
        connect(leaf, top)
    print(root_up.top())
    print(list(root_up.leaves()))

    for node in root_up.iterate():
        print(node)

    for node in root_up.iterate_as_tree():
        print(node)


if __name__ == '__main__':
    test_dag_iteration()


import msgpack
import os
from utils_ak.granular_storage.table.timeseries.fs_basic import BasicFsTimeSeriesTable


class FsMsgPackTimeSeriesTable(BasicFsTimeSeriesTable):
    def __init__(self, fn, encoding='utf-8'):
        super().__init__(fn)
        self.fn = fn
        self.encoding = encoding

    def encode(self, msg):
        return msgpack.packb(msg, use_bin_type=True)

    def read(self):
        if not os.path.exists(self.fn):
            raise StopIteration

        with open(self.fn, 'rb') as f:
            for unpacked in msgpack.Unpacker(f, raw=False):
                yield unpacked


if __name__ == '__main__':
    from datetime import datetime

    storage = FsMsgPackTimeSeriesTable('tmp2.tss')

    for i in range(3):
        storage.store(datetime.now().timestamp(), 'key', {'a': i})

    # flush
    storage.close()
    for value in storage.read():
        print(1, value)
    print()

    storage.store_many([[datetime.now().timestamp(), 'key', {'b': i}] for i in range(3)])

    # flush
    storage.close()

    for value in storage.read():
        print(2, value)
    print()

    storage.store_many([[datetime.now().timestamp(), 'key', {'c': i}] for i in range(3)], safe=True, overwrite=True)

    # already flushed!

    for value in storage.read():
        print(3, value)
    print()
    storage.clear()



from utils_ak.fluid_flow import *



def test_flow_hub_1():
    parent = Container('Parent', value=100, max_pressures=[None, 20])

    hub = Hub('Hub')

    child1 = Container('Child1', max_pressures=[15, None])
    child2 = Container('Child2', max_pressures=[10, None])

    pipe_connect(parent, hub, 'parent-hub')
    pipe_connect(hub, child1, 'hub-child1')
    pipe_connect(hub, child2, 'hub-child2')

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_hub_2():
    parent = Container('Parent', value=100, max_pressures=[None, 20])

    hub = Hub('Hub')

    child1 = Processor('Child1', max_pressures=[15, None], limits=[30, None])
    child2 = Processor('Child2', max_pressures=[10, None])

    pipe_connect(parent, hub, 'parent-hub')
    pipe_connect(hub, child1, 'hub-child1')
    pipe_connect(hub, child2, 'hub-child2')

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_hub_3():
    parent = Container('Parent', item='a', value=100, max_pressures=[None, 20])

    hub = Hub('Hub')

    child1 = Container('Child1', item='b', max_pressures=[15, None])
    child2 = Container('Child2', item='a', max_pressures=[10, None])

    pipe_connect(parent, hub, 'parent-hub')
    pipe_connect(hub, child1, 'hub-child1')
    pipe_connect(hub, child2, 'hub-child2')

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


if __name__ == '__main__':
    # test_flow_hub_1()
    # test_flow_hub_2()
    test_flow_hub_3()


#!/usr/bin/env python3
import logging
import git
import os

from utils_ak.log import configure_stream_logging

configure_stream_logging()

git_root_path = os.path.abspath(os.path.dirname(__file__) + '/../../..')
gits = ['utils_ak', 'barsim', 'exchanges', 'prod']

logging.info('Cloning repos...')
# 1) clone all gits
for git_repo in gits:
    git_path = os.path.join(git_root_path, git_repo)
    if not os.path.exists(git_path):
        logging.info(f'Cloning: {git_repo}')
        git.Repo.clone_from('git@bitbucket.org:qtb/{}.git'.format(git_repo), git_path)
logging.info('Done.')

logging.info('Pulling repos...')
# 2) pull all gits
from utils_ak.git import git_pull_many
from pprint import pprint

pprint(git_pull_many(git_root_path, repos=gits))
logging.info('Done.')



import time
from loguru import logger
from utils_ak.simple_microservice import run_listener_async


def test_worker(worker_cls, payload, message_broker):
    from utils_ak.loguru import configure_loguru_stdout
    configure_loguru_stdout('INFO')
    run_listener_async('monitor', message_broker=message_broker)
    time.sleep(2)
    worker = worker_cls('worker_id', payload, message_broker=message_broker)
    worker.run()
    logger.info('Finished batch')


def test_worker_deployment(deployment_fn, message_broker):
    from utils_ak.loguru import configure_loguru_stdout
    configure_loguru_stdout('DEBUG')
    run_listener_async('monitor', message_broker=message_broker)

    from utils_ak.deployment import DockerController
    ctrl = DockerController()

    import anyconfig
    deployment = anyconfig.load(deployment_fn)

    ctrl.stop(deployment['id'])
    ctrl.start(deployment)
    time.sleep(5)
    ctrl.stop(deployment['id'])



import subprocess
import os

from utils_ak.str import cast_unicode


def execute(cmd, async_=False):
    if async_:
        process = subprocess.Popen(cmd, shell=True)
    else:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        output, error = process.communicate()
        return ' '.join([cast_unicode(x) for x in [output, error] if x]), process.returncode


class AwsDumper:
    def __init__(self, local_file_path, bucket, remote_file_dir,  *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.local_file_path = local_file_path
        self.remote_file_dir = remote_file_dir
        self.bucket = bucket

    def tick(self):
        local_file_dir, _ = os.path.split(self.local_file_path)
        message, code = execute(
            f'aws s3 sync {local_file_dir} s3://{self.bucket}/{self.remote_file_dir} '
            f'--exclude="*" --include={self.local_file_path} --delete')

        if code != 0:
            raise Exception(message)

        return message



import subprocess
from loguru import logger

from utils_ak.str import cast_unicode


def execute(cmd, is_async=False, verbose=True):
    if verbose:
        logger.debug('Executing command', cmd=cmd, is_async=is_async)
    if is_async:
        subprocess.Popen(cmd, shell=True)
    else:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        output, error = process.communicate()
        res = ' '.join([cast_unicode(x) for x in [output, error] if x])
        if verbose:
            logger.debug('Execution response', response=res)
        return res


import logging
import os
import sys
import time
import traceback
import asyncio

from utils_ak.callback_timer import CallbackTimer, ScheduleTimer, CallbackTimers
from utils_ak.architecture.func import PrefixHandler
from utils_ak.str import cast_unicode
from utils_ak.serialization import JsonSerializer
from utils_ak.message_queue import cast_message_broker
from utils_ak.loguru import patch_trace

from loguru import logger as global_logger


TIME_EPS = 0.001


class SimpleMicroservice(object):
    """ Microservice base class with timers and subscriber. Works on asyncio. """

    def __init__(self, id, message_broker, logger=None, serializer=None):
        self.id = id

        # aio
        self.tasks = []

        # sync
        self.timers = []

        # {collection: callback}
        self.callbacks = {}

        self.broker = cast_message_broker(message_broker)

        # {broker: f'{collection}::{topic}'}
        self.subscribed_to = {}

        self.logger = logger or global_logger
        self.logger = self.logger.bind(inner_source=str(id))
        self.logger = self.logger.patch(patch_trace)

        self.default_exception_timeout = 10.
        self.max_exception_timeout = 3600
        self.fail_count = 0

        self.is_active = True

        self.serializer = serializer or JsonSerializer()

    def stop(self):
        self.is_active = False

    def _args_formatter(self, topic, msg):
        return (cast_unicode(topic), self.serializer.decode(msg)), {}

    def add_timer(self, *args, **kwargs):
        """
        :param interval: timer will be run every interval
        :param callback: func
        """
        self.timers.append(CallbackTimer(*args, **kwargs))

    def add_schedule(self, *args, **kwargs):
        """
        :param pattern: cron-like pattern with seconds:
              sec min hour monthday month weekday
              *    5    *     *       *     *     -> run every 5th minute of every hour
        :param callback:
        :param init_run: bool
        """
        self.timers.append(ScheduleTimer(*args, **kwargs))

    def subscribe(self, collection, topic):
        self.broker.subscribe(collection, topic)

    def add_callback(self, collection, topic, callback=None, formatter='default', filter=None, topic_formatter=cast_unicode):
        self.broker.subscribe(collection, topic)
        self._add_callback(collection, topic, callback, formatter, filter, topic_formatter)

    def _add_callback(self, collection, topic, callback=None,
                      formatter='default', filter=None, topic_formatter=cast_unicode):
        assert type(topic) == str, 'Topic must be str'

        if formatter == 'default':
            formatter = self._args_formatter

        handler = self.callbacks.setdefault(collection, PrefixHandler())
        handler.set_topic_formatter(topic_formatter)
        handler.add(topic, callback=callback, filter=filter, formatter=formatter)
        return handler

    def publish(self, collection, topic, msg):
        self.broker.publish(collection, topic, msg)

    def publish_json(self, collection, topic, msg):
        self.publish(collection, topic, self.serializer.encode(msg))

    def wrap_coroutine_timer(self, timer):
        async def f():
            while True:
                try:
                    has_ran = await timer.run_if_possible_aio()

                    if has_ran and self.fail_count != 0:
                        self.logger.debug('Success. Resetting the failure counter')
                        self.fail_count = 0

                except Exception as e:
                    self.on_exception(e, 'Exception occurred at the timer callback')

                if not self.is_active:
                    return

                await asyncio.sleep(max(timer.next_call - time.time() + TIME_EPS, 0))

        return f()

    def wrap_broker_coroutine(self, timeout=0.01):
        async_supported = self.broker.async_supported

        async def f():
            while True:
                try:
                    if async_supported:
                        received = await self.broker.aiopoll(timeout=1.)
                    else:
                        received = self.broker.poll(timeout)

                    if received:
                        collection, topic, msg = received
                        try:
                            self.logger.info(f'Received new message', custom={'topic': str(topic), 'msg': str(msg)})
                            await self.callbacks[collection].aiocall(topic, msg)

                            if self.fail_count != 0:
                                self.logger.info('Success. Resetting the failure counter')
                                self.fail_count = 0

                        except Exception as e:
                            self.on_exception(e, f"Exception occurred at the callback")
                except Exception as e:
                    self.on_exception(e, f'Failed to receive the message')

                if not self.is_active:
                    return

                await asyncio.sleep(0)

        return f()

    def _aiorun(self):
        self.loop = asyncio.get_event_loop()
        # self.loop.set_debug(True)
        self.logger.info('Microservice started')

        for timer in self.timers:
            self.tasks.append(self.wrap_coroutine_timer(timer))

        self.tasks.append(self.wrap_broker_coroutine())

        self.tasks = [asyncio.ensure_future(task) for task in self.tasks]
        self.loop.run_until_complete(asyncio.wait(self.tasks))

    def _run(self, timeout=0.01):
        self.logger.info('Microservice started')

        self.callback_timers = CallbackTimers()
        for timer in self.timers:
            self.callback_timers.add_timer(timer)

        while True:
            success = False
            try:
                if len(self.timers) > 0:
                    if time.time() > self.callback_timers.next_call:
                        # some timer is ready
                        for timer in self.callback_timers.timers:
                            try:
                                timer.run_if_possible()
                            except Exception as e:
                                self.on_exception(e, 'Exception occurred at the timer callback')
                            else:
                                success = True

                try:
                    received = self.broker.poll(timeout)
                    if not received:
                        continue
                    collection, topic, msg = received
                    try:
                        self.logger.debug(f'Received new message', topic=topic, msg=msg)
                        self.callbacks[collection].call(topic, msg)
                    except Exception as e:
                        self.on_exception(e, f"Exception occurred")
                    else:
                        success = True

                except Exception as e:
                    self.on_exception(e, 'Failed to receive the message')

                if not self.is_active:
                    self.logger.info('Microservice not active. Stopping')
                    break

            except Exception as e:
                self.on_exception(e, 'Global exception occurred')

            if success and self.fail_count != 0:
                self.logger.info('Success. Resetting the failure counter')
                self.fail_count = 0

    def run(self, asyncio=True):
        if asyncio:
            return self._aiorun()
        else:
            return self._run()

    def on_exception(self, e, msg):
        self.logger.error('Generic microservice error')
        to_sleep = min(self.default_exception_timeout * 2 ** (self.fail_count - 1), self.max_exception_timeout)
        time.sleep(to_sleep)
        self.fail_count += 1



import os
from utils_ak.granular_storage.df.file.stream import StreamFile
import pandas as pd


# todo: add test for extensability
# todo: debug


class FeatherFile(StreamFile):
    def __init__(self, fn, buffer_size=None, buffer_timeout=60, nthreads=1):
        self.fn = fn
        self.nthreads = nthreads
        super().__init__(buffer_size, buffer_timeout)

    def append(self, df, append=True):
        if append and os.path.exists(self.fn):
            old_df = self.read()
            df = pd.concat([old_df, df], axis=0)
        df.reset_index().to_feather(self.fn)

    def append_stream(self, df):
        super().append_stream(df, 'default')

    def read(self):
        return pd.read_feather(self.fn, self.nthreads).set_index('index')



import urllib
import json
import os
import re

from utils_ak.os import remove_path, execute


def get_current_notebook_path():
    from notebook import notebookapp
    import ipykernel

    """Returns the absolute path of the Notebook or None if it cannot be determined
    NOTE: works only when the security is token-based or there is also no password
    """
    connection_file = os.path.basename(ipykernel.get_connection_file())
    kernel_id = connection_file.split('-', 1)[1].split('.')[0]

    for srv in notebookapp.list_running_servers():
        try:
            if srv['token'] =='' and not srv['password']:  # No token and no password, ahem...
                req = urllib.request.urlopen(srv['url']+'api/sessions')
            else:
                req = urllib.request.urlopen(srv['url']+'api/sessions?token='+srv['token'])
            sessions = json.load(req)
            for sess in sessions:
                if sess['kernel']['id'] == kernel_id:
                    return os.path.join(srv['notebook_dir'],sess['notebook']['path'])
        except:
            pass  # There may be stale entries in the runtime directory
    return None


def get_current_notebook_code():
    path = get_current_notebook_path()
    path = path.replace('\\', '/')
    s = fr'jupyter nbconvert --to script "{path}"'
    execute(s)
    with open(path.replace('.ipynb', '.py'), 'r') as f:
        code = f.read()
    remove_path(path.replace('.ipynb', '.py'))
    return code


def parse_cells(notebook_code):
    cell_headers = re.findall(r'(In\[[\d\s]+\])', notebook_code)
    for cell_header in cell_headers:
        notebook_code = re.sub(cell_header.replace('[', '\[').replace(']', '\]'), 'In[]', notebook_code, 1)

    res = ''
    for line in notebook_code.split('\n'):
        if 'In[]' in line:
            res += '>split_cell'
        else:
            res += line
            res += '\n'
    cells = res.split('>split_cell')

    return cells[1:]


def is_in_ipynb():
    try:
        cfg = get_ipython().config
        if cfg['IPKernelApp']['parent_appname'] == 'ipython-notebook':
            return True
        else:
            return False
    except NameError:
        return False


def safe_display(obj):
    if is_in_ipynb():
        from IPython.display import display
        display(obj)
    else:
        print(obj)


def safe_markdown(obj):
    if is_in_ipynb():
        from IPython.display import Markdown
        Markdown(obj)
    else:
        print(obj)


import os
import time
import fire
from loguru import logger


def main(name=None, run_forever=True):
    name = name or os.environ.get('NAME') or 'World'
    logger.info(f'Hello {name}!')
    if run_forever:
        while True:
            time.sleep(1)


if __name__ == '__main__':
    fire.Fire(main)


from utils_ak.fluid_flow import *
from utils_ak.block_tree import *

import warnings
warnings.filterwarnings('ignore')


def test_water_line_flow_1():
    drenator = Container('Drenator', value=1000, max_pressures=[None, None])

    melting1 = Processor('Melting1', items=['a', 'a'], max_pressures=[1000, 1000], processing_time=0, limits=[750, 750])
    melting2 = Processor('Melting2', items=['b', 'b'], max_pressures=[1000, 1000], processing_time=0, limits=[250, 250])
    melting_queue = Queue('MeltingQueue', [melting1, melting2], break_funcs={'in': lambda old, new: 1})

    cooling1 = Processor('Cooling1', items=['a', 'a'], max_pressures=[None, None], processing_time=0.5, limits=[750, 750])
    cooling2 = Processor('Cooling2', items=['b', 'b'], max_pressures=[None, None], processing_time=0.5, limits=[250, 250])
    cooling_queue = Queue('CoolingQueue', [cooling1, cooling2])

    packing_hub = Hub('Hub')

    packing1 = Processor('Packing1', items=['a', 'a1'], max_pressures=[200, None], processing_time=0, limits=[750, None])
    packing2 = Processor('Packing2', items=['b', 'b1'], max_pressures=[400, None], processing_time=0, limits=[100, None])
    packing_queue1 = Queue('PackingQueue1', [packing1, packing2])

    # using sequence as packing
    container = Container('Container3', item='b', max_pressures=[200, None], limits=[150, None])
    # slow packing...
    processor = Processor('Processor3', items=['b', 'b2'], max_pressures=[50, None], processing_time=0)
    packing3 = Sequence('Packing3', [container, processor])
    packing_queue2 = Queue('PackingQueue1', [packing3])

    pipe_connect(drenator, melting_queue, 'drenator-melting')
    pipe_connect(melting_queue, cooling_queue, 'melting-cooling')
    pipe_connect(cooling_queue, packing_hub, 'cooling-hub')

    pipe_connect(packing_hub, packing_queue1, 'hub-packing_queue1')
    pipe_connect(packing_hub, packing_queue2, 'hub-packing_queue2')

    flow = FluidFlow(drenator, verbose=True)
    run_flow(flow)

    maker, make = init_block_maker('root', axis=1)

    for node in drenator.iterate('down'):
        if node.active_periods():
            for period in node.active_periods():
                label = '-'.join([str(node.name), period[0]])
                beg, end = period[1:]
                beg, end = custom_round(beg * 60, 5) // 5, custom_round(end * 60, 5) // 5
                make(label, x=[beg, 0], size=(end - beg, 1))
    print(maker.root.tabular())


def test_water_line_flow_2_multiple_boilings():
    n_boilings = 2
    packing1_1 = Processor('Packing1', items=['a', 'a1'], max_pressures=[200, None], processing_time=0, limits=[750, None])
    packing2_1 = Processor('Packing2', items=['b', 'b1'], max_pressures=[400, None], processing_time=0, limits=[100, None])
    packing1_2 = Processor('Packing1', items=['a', 'a1'], max_pressures=[200, None], processing_time=0, limits=[750, None])
    packing2_2 = Processor('Packing2', items=['b', 'b1'], max_pressures=[400, None], processing_time=0, limits=[100, None])
    packing_queue1 = Queue('PackingQueue1', [packing1_1, packing2_1, packing1_2, packing2_2])

    packing3_1 = Processor('Packing3', items=['b', 'b2'], max_pressures=[200, None], processing_time=0, limits=[150, None])
    packing3_2 = Processor('Packing3', items=['b', 'b2'], max_pressures=[200, None], processing_time=0, limits=[150, None])
    packing_queue2 = Queue('PackingQueue1', [packing3_1, packing3_2])

    for _ in range(n_boilings):
        drenator = Container('Drenator', value=1000, max_pressures=[None, None])

        melting1 = Processor('Melting1', items=['a', 'a'], max_pressures=[1000, 1000], processing_time=0, limits=[750, 750])
        melting2 = Processor('Melting2', items=['b', 'b'], max_pressures=[1000, 1000], processing_time=0, limits=[250, 250])
        melting_queue = Queue('MeltingQueue', [melting1, melting2], break_funcs={'in': lambda old, new: 1})

        cooling1 = Processor('Cooling1', items=['a', 'a'], max_pressures=[None, None], processing_time=0.5, limits=[750, 750])
        cooling2 = Processor('Cooling2', items=['b', 'b'], max_pressures=[None, None], processing_time=0.5, limits=[250, 250])
        cooling_queue = Queue('CoolingQueue', [cooling1, cooling2])

        packing_hub = Hub('Hub')

        pipe_connect(drenator, melting_queue, 'drenator-melting')
        pipe_connect(melting_queue, cooling_queue, 'melting-cooling')
        pipe_connect(cooling_queue, packing_hub, 'cooling-hub')

        pipe_connect(packing_hub, packing_queue1, 'hub-packing_queue1')
        pipe_connect(packing_hub, packing_queue2, 'hub-packing_queue2')

        flow = FluidFlow(drenator, verbose=True)
        run_flow(flow)

        maker, make = init_block_maker('root', axis=1)

        for node in drenator.iterate('down'):
            if node.active_periods():
                for period in node.active_periods():
                    label = '-'.join([str(node.name), period[0]])
                    beg, end = period[1:]
                    beg, end = custom_round(beg * 60, 5) // 5, custom_round(end * 60, 5) // 5
                    make(label, x=[beg, 0], size=(end - beg, 1))
        print(maker.root.tabular())

        # clean-up
        for node in drenator.iterate('down'):
            node.reset()

        # remove packers from current boiling
        pipe_disconnect(packing_hub, packing_queue1)
        pipe_disconnect(packing_hub, packing_queue2)


if __name__ == '__main__':
    test_water_line_flow_1()
    # test_water_line_flow_2_multiple_boilings()


from utils_ak.clock import *
from utils_ak.fluid_flow.actor import Actor
from utils_ak.fluid_flow.actors.pipe import PipeMixin
from utils_ak.fluid_flow.calculations import *
import pandas as pd


class Container(Actor, PipeMixin):
    def __init__(self, name, value=0, item='default', max_pressures=None, limits=None):
        super().__init__(name)
        self.item = item
        self.value = value

        self.df = pd.DataFrame(index=['in', 'out'], columns=['max_pressure', 'limit', 'collected'])
        self.df['max_pressure'] = max_pressures
        self.df['limit'] = limits
        self.df['collected'] = 0.

        self.transactions = []

    def is_limit_reached(self, orient):
        if self.df.at[orient, 'limit'] and abs(self.df.at[orient, 'collected'] - self.df.at[orient, 'limit']) < ERROR:
            return True
        return False

    def update_value(self, ts, factor=1):
        if self.last_ts is None:
            return

        self.add_value(ts, 'in', (ts - self.last_ts) * self.speed('in') * factor)
        self.add_value(ts, 'out', -(ts - self.last_ts) * self.speed('out'))

    def add_value(self, ts, orient, value):
        if not value:
            return
        self.value += value
        self.df.at[orient, 'collected'] += abs(value)
        self.transactions.append([self.last_ts, ts, value])

    def active_periods(self, orient='in'):
        if not self.transactions:
            return []
        return [[self.item, self.transactions[0][0], self.transactions[-1][1]]]

    def update_pressure(self, ts, orients=('in', 'out')):
        for orient in orients:
            if self.pipe(orient):
                pressure = self.df.at[orient, 'max_pressure'] if not self.is_limit_reached(orient) else 0
                self.pipe(orient).set_pressure(orient, pressure, self.item)

    def update_speed(self, ts, set_out_pressure=True):
        input_speed = self.speed('in')

        if set_out_pressure:
            if self.pipe('out') and abs(self.value) < ERROR:
                self.pipe('out').set_pressure('out', nanmin([self.pipe('out').pressures['out'], input_speed]), self.item)

    def reset(self):
        super().reset()
        self.transactions = []

    def update_triggers(self, ts):
        values = []
        if self.drain() < 0:
            values.append(['empty_container', self.value, self.drain()])

        for orient in ['in', 'out']:
            if self.df.at[orient, 'limit']:
                values.append([f'{orient} limit', self.df.at[orient, 'limit'] - self.df.at[orient, 'collected'], self.speed(orient)])

        values = [value for value in values if value[1] > ERROR and abs(value[2]) > ERROR]
        etas = [value[1] / abs(value[2]) for value in values]
        if etas:
            self.add_event('update.trigger', ts + min(etas), {})

    def __str__(self):
        return f'Container {self.name}:{self.item}'

    def stats(self):
        return {'value': self.value}

    def display_stats(self):
        return self.value



from utils_ak.time.dt import *
from utils_ak.pd_tools import find_row
from utils_ak.os_tools import remove

from utils_ak.tqdm_tools import tqdm_ak


class GranularSync(object):
    def __init__(self, gs_from, gs_to):
        self.gs_from = gs_from
        self.gs_to = gs_to

    def sync(self, path_from, path_to, flag_migration=False):
        if not self.gs_from.exists(path_from):
            raise Exception(f'Path not exists {path_from} in gran_from')

        gran_from = self.gs_from.get_granular(path_from)
        if not self.gs_to.exists(path_to):
            # inherit format from gran_from
            gran_to = self.gs_to.get_granular(path_to, level=gran_from.level, pattern=gran_from.pattern)
        else:
            gran_to = self.gs_to.get_granular(path_to)

        # sync gran1 into gran2
        struct_from = gran_from.get_struct_df()
        struct_to = gran_to.get_struct_df()

        last_struct_to_df = struct_to.groupby('key').agg('last')
        last_rows = last_struct_to_df.to_dict('index')

        for key in tqdm_ak(struct_from['key'].unique()):
            if key not in last_rows:
                # key is not present in gran_to
                df = gran_from.read(key)
            else:
                last_df_to = gran_to.read(key, beg=last_rows[key]['ts'])
                last_row = last_df_to.iloc[[-1]]

                df = gran_from.read(key, beg=last_row.index[0])

                last_row_idx = find_row(df, last_row)

                if last_row_idx is not None:
                    df = df.iloc[last_row_idx + 1:]

                if len(df) == 0:
                    continue
            gran_to.append(df, key)

        # see Streamer
        if flag_migration:
            gran_to.upd_metadata({'migration_complete': True})


if __name__ == '__main__':
    from utils_ak.granular_storage.df.gran_df import GranularFeatherStorage

    gs1 = GranularFeatherStorage('gs_root1')
    gs2 = GranularFeatherStorage('gs_root2')

    index = [cast_datetime('2018.05.01') + timedelta(hours=6) * 0 for i in range(6)]
    # df1 = pd.DataFrame(np.random.randn(6, 1), index=index, columns=['0'])
    df1 = pd.DataFrame([np.random.randn(1, 1)[0]] * 5 + [[2]], index=index, columns=['0'])

    gran1 = gs1.get_granular('gran1', level=0, mode='w')
    gran2 = gs2.get_granular('gran2', level=4, mode='w')

    gran1.append(df1, 'foo1.feather')
    gran2.append(df1.iloc[:-1], 'foo1.feather')

    sync = GranularSync(gs1, gs2)
    sync.sync('gran1', 'gran2')
    gran2 = gs2['gran2']

    print((gran1['foo1.feather'].equals(gran2['foo1.feather'])))

    # clean test granular_storage
    remove('gs_root1/')
    remove('gs_root2/')



import asyncio
from datetime import datetime
import numpy as np
import logging

from utils_ak.zmq import endpoint
from utils_ak.simple_microservice import SimpleMicroservice, run_listener_async
logging.basicConfig(level=logging.INFO)

ping_logger = logging.getLogger('ping')
pong_logger = logging.getLogger('pong')


class Ping(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__('Test publisher', logger=ping_logger, *args, **kwargs)
        self.add_callback('pong', '', self.send_ping)
        self.add_timer(self.print_random, 2.0)

    async def print_random(self):
        timeout = np.random.uniform(5.0, 6.0)
        self.logger.info(f'Stop: {datetime.now()} {timeout}')
        await asyncio.sleep(timeout)
        self.logger.info(f'Resume: {datetime.now()} {timeout}')

    async def send_ping(self, topic, msg):
        self.logger.info(f'Received {topic} {msg}')
        await asyncio.sleep(1.0)
        self.publish_json('ping', '', {'msg': 'ping'})


class Pong(SimpleMicroservice):
    def __init__(self, *args, **kwargs):
        super().__init__('Test publisher', logger=pong_logger, *args, **kwargs)
        self.add_callback('ping', '', self.send_pong)

    async def send_pong(self, topic, msg):
        self.logger.info(f'Received {topic} {msg}')
        await asyncio.sleep(1.0)
        self.publish_json('pong', '', {'msg': 'pong'})


def run_ping():
    ping = Ping(message_broker=('zmq',  {'endpoints': {'ping': {'type': 'sub', 'endpoint': endpoint('localhost', 6554)},
                                        'pong': {'type': 'sub', 'endpoint': endpoint('localhost', 6555)}}}))

    async def send_initial():
        await asyncio.sleep(1.0)
        await ping.send_ping('init', 'init')

    ping.tasks.append(asyncio.ensure_future(send_initial()))
    ping.run()


def run_pong():
    Pong(message_broker=('zmq',  {'endpoints': {'ping': {'type': 'sub', 'endpoint': endpoint('localhost', 6554)},
                                        'pong': {'type': 'sub', 'endpoint': endpoint('localhost', 6555)}}})).run()


if __name__ == '__main__':
    import multiprocessing
    multiprocessing.Process(target=run_ping).start()
    multiprocessing.Process(target=run_pong).start()



from utils_ak.architecture import TopicHandler, PrefixHandler, Handler

if __name__ == '__main__':
    def callback1(msg):
        print('callback1', msg)
        return 1


    def callback2(msg):
        print('callback2', msg)
        return 2


    def filter(msg):
        print('filter', msg)
        return 'bar1' in msg


    def formatter(msg):
        return (msg.split('.')[0],), {}


    handler = Handler(callback=callback1)
    print('Test 1')
    print(handler('foo.bar1'))
    print(handler('foo.bar2'))
    print()

    handler = Handler(callback=[callback1, callback2], filter=filter)
    print('Test 2')
    print(handler('foo.bar1'))
    print(handler('foo.bar2'))
    print()

    handler = Handler(callback=[callback1, callback2])
    handler.add_filter(filter, rule=lambda b: not b)
    print('Test 3')
    print(handler('foo.bar1'))
    print(handler('foo.bar2'))
    print()

    handler = Handler(callback=[callback1, callback2], filter=filter, formatter=formatter)
    print('Test 4')
    print(handler('foo.bar1'))
    print(handler('foo.bar2'))
    print()


    def topic_callback1(topic, msg):
        print('callback1', topic, msg)
        return 1

    print('Test 5. Topic Handler')
    handler = TopicHandler()
    handler.add('a.b', callback=topic_callback1)
    handler('a', '')
    handler('b', '')
    handler('a.b', '')
    print()

    print('Test 6. Prefix Handler')
    handler = PrefixHandler()
    handler.add('a', callback=topic_callback1)
    handler('a', '')
    handler('b', '')
    handler('a.b', '')
    print()

    # aio stuff
    import asyncio

    async def aiocallback(topic='', msg=''):
        await asyncio.sleep(0.5)
        print('aiocallback', topic, msg)
        return 3

    loop = asyncio.get_event_loop()

    print('Test 7. AIO 1')
    handler = Handler(callback=[callback1, callback2, aiocallback], filter=filter)
    loop.run_until_complete(handler.aiocall('foo.bar1'))
    loop.run_until_complete(handler.aiocall('foo.bar2'))
    print()

    print('Test 7. AIO Topic Handler')
    handler = TopicHandler()
    handler.add('a.b', callback=aiocallback)
    loop.run_until_complete(handler.aiocall('a', ''))
    loop.run_until_complete(handler.aiocall('a', ''))
    loop.run_until_complete(handler.aiocall('a.b', ''))

    print()

    print('Test 7. AIO Prefix Handler')
    handler = PrefixHandler()
    handler.add('a', callback=[topic_callback1, aiocallback])
    loop.run_until_complete(handler.aiocall('a', ''))
    loop.run_until_complete(handler.aiocall('b', ''))
    loop.run_until_complete(handler.aiocall('a.b', ''))
    print()



""" Git functionality. """
import glob
import os
import sys
from collections import defaultdict

import git
import traceback

from utils_ak.time.dt import cast_datetime
from utils_ak.re import search_one


def is_git_repo(path):
    try:
        _ = git.Repo(path).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False


def add_git_repos_to_path(path, recursive=False):
    if not recursive:
        paths = glob.glob(os.path.join(path, '*/'))
    else:
        paths = glob.glob(os.path.join(path, '**/*/'), recursive=True)

    paths = [path for path in paths if is_git_repo(path)]

    for path in paths:
        sys.path.append(path)


def git_pull(path, full_output=False):
    if not is_git_repo(path):
        return {'status': 'fail', 'e': 'Not a git repo'}

    res = defaultdict(str)
    try:
        g = git.cmd.Git(path)
        # if stash:
        #     stash_output = g.stash()
        #     res['output'] += stash_output + '\n'

        output = g.pull()
        res['output'] += output + '\n'

        # if stash:
        #     try:
        #         unstash_output = g.stash('apply')
        #         res['output'] += unstash_output + '\n'
        #     except:
        #         pass

        if 'Already up-to-date' in output:
            res['status'] = 'Already up-to-date'
        else:
            res['status'] = 'success'

    except Exception as e:
        res['status'] = 'fail'
        res['e'] = traceback.format_exc()
    res = dict(res)
    if not full_output:
        res = res['status']
    return res


def git_pull_many(path, repos=None, recursive=False):
    if not recursive:
        paths = glob.glob(os.path.join(path, '*/'))
    else:
        paths = glob.glob(os.path.join(path, '**/*/'), recursive=True)
    paths = [path for path in paths if is_git_repo(path)]
    if repos:
        paths = [path for path in paths if os.path.basename(os.path.dirname(path)) in repos]
    return {path: git_pull(path) for path in paths}


def get_commit_info(path):
    repo = git.Repo(path)
    commit_info = repo.git.show()

    commit_dic = {'hash': search_one(r'commit (.+)\n', commit_info),
                  'date': search_one(r'Date:\s+(.+)\n', commit_info),
                  'author': search_one(r'Author:\s+(.+)\n', commit_info),
                  'full_info': commit_info}
    try:
        commit_dic['date'] = cast_datetime(commit_dic['date'])
    except:
        pass
    return commit_dic


if __name__ == '__main__':
    # res = git_pull_many(r'C:\Users\xiaomi\YandexDisk\IT\Python\pycharm')
    # res = git_pull_many('/home/jentos/Projects/quantribution')
    # pprint(res)
    print(get_commit_info(r'C:\Users\xiaomi\YandexDisk\IT\Python\pycharm\prod'))



from utils_ak.fluid_flow import *


def test_flow_sequence_1():
    c1 = Container('Input', value=100, max_pressures=[None, 10])

    sequence = Sequence('Sequence', containers=[Container(str(i), max_pressures=[5, None]) for i in range(10)])

    c2 = Container('Ouput', max_pressures=[None, None])

    pipe_connect(c1, sequence)
    pipe_connect(sequence, c2)
    flow = FluidFlow(c1, verbose=True)
    run_flow(flow)


def test_flow_sequence_2():
    c1 = Container('Input', value=100, max_pressures=[None, 10])

    sequence = Sequence('Sequence', containers=[Container(str(i), max_pressures=[2 - i, None]) for i in range(2)])

    c2 = Container('Ouput', max_pressures=[None, None])

    pipe_connect(c1, sequence)
    pipe_connect(sequence, c2)
    flow = FluidFlow(c1, verbose=True)
    run_flow(flow)


if __name__ == '__main__':
    # test_flow_sequence_1()
    test_flow_sequence_2()


import os
import pickledb

from utils_ak.state.provider.provider import StateProvider
from utils_ak.os import make_directories


class PickleDBStateProvider(StateProvider):
    """ I keep a number (which is a called a state). I allow for the state to not change in case of crash. """

    def __init__(self, fn, state_key='state', default_state=None):
        super().__init__()

        self.state_key = state_key
        self.fn = fn
        make_directories(self.fn)
        # crash canary
        self.canary_name = fn + ".canary"

        self.db = pickledb.load(fn, False)

        self.default_state = default_state or {}
        self.state = None
        self._recover()

    def _recover(self):
        # Check if canary is there
        if os.path.isfile(self.canary_name):
            # We have crashed while updating state, recovering the state
            self.canary_db = pickledb.load(self.canary_name, False)
            self.set_state(self.canary_db.get(self.state_key) or self.default_state)
            os.remove(self.canary_name)

    def get_state(self):
        if self.state is None:
            self.state = self.db.get(self.state_key) or self.default_state
        return self.state

    def set_state(self, state):
        self.state = state
        self.db.set(self.state_key, state)
        self.db.dump()  # Dumping instantly

    def __enter__(self):
        self.canary_db = pickledb.load(self.canary_name, False)
        self.canary_db.set(self.state_key, self.get_state())
        self.canary_db.dump()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.remove(self.canary_name)


if __name__ == "__main__":
    storage = PickleDBStateProvider('state.pickledb')
    storage.set_state(16)
    assert storage.get_state() == 16

    PickleDBStateProvider('state.pickledb').set_state(17)
    assert PickleDBStateProvider('state.pickledb').get_state() == 17

    with PickleDBStateProvider('state.pickledb') as sp:
        assert sp.get_state() == 17
        sp.set_state(18)
        assert sp.get_state() == 18
    print(storage.get_state())



import psutil

import subprocess
from utils_ak.str import cast_unicode

import re

import pandas as pd


def execute(cmd, run_async=False):
    if run_async:
        process = subprocess.Popen(cmd, shell=True)
    else:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        output, error = process.communicate()
        return ' '.join([cast_unicode(x) for x in [output, error] if x]), process.returncode


def get_system_usage_stats():
    # todo: add cpu stats
    stats = {}

    # svmem(total=17058336768, available=6641549312, percent=61.1, used=10416787456, free=6641549312)
    memory = psutil.virtual_memory()
    stats['ram'] = {'total': memory.total, 'available': memory.available, 'free': memory.free,
                    'percent': memory.percent}

    # sdiskusage(total=254277836800, used=212004134912, free=42273701888, percent=83.4)
    disk = psutil.disk_usage('/')
    stats['disk'] = {'total': disk.total, 'used': disk.used, 'free': disk.free, 'percent': disk.percent}
    return stats


def get_dfh():
    output = execute("df -h | grep -v 'map '")  # grep -v 'map ' for mac book
    s = output[0]
    s = s.replace('Mounted on', 'Mounted')
    s = s.replace('%iused', 'Use%')  # for mac
    lines = s.strip().split('\n')
    lines = [re.sub('\s+', ' ', line) for line in lines]
    values = [line.split(' ') for line in lines]
    df = pd.DataFrame(values[1:], columns=values[0])
    return df



from matplotlib import colors


def cast_color(obj):
    if isinstance(obj, (tuple, list)):
        if all(isinstance(x, int) for x in obj):
            return '#{:02x}{:02x}{:02x}'.format(*obj)
        elif all(isinstance(x, float) for x in obj):
            return colors.to_hex(obj)
    elif isinstance(obj, str):
        try:
            return cast_color(colors.to_rgb(obj))
        except:
            pass
    raise Exception('Unknown object type')


if __name__ == '__main__':
    print(cast_color('grey'))
    print(cast_color((1., 0., 1.)))
    print(cast_color((15, 15, 16)))
    print(cast_color('#808081'))


import os
import re
import pandas as pd
import json

from datetime import datetime, timedelta

from utils_ak.granular_storage.df.file import CSVFile, FeatherFile, MsgPackFile, ParquetFile
from utils_ak.granular_storage.df.gran_df.enums import *

from utils_ak.time.dt import cast_timedelta, cast_datetime
from utils_ak.os import list_files, make_directories, remove_path
from utils_ak.pandas import merge
from utils_ak.re import re as re_tools


# todo: optimize upd_struct usage. Minimize its usage
# todo: parse level and pattern without metadata presetn


class NoFilesException(Exception):
    pass


class Granular(object):
    default_patterns = {LEVEL_ZERO: '{key}',
                        LEVEL_YEAR: '%Y/%Y_{key}',
                        LEVEL_MONTH: '%Y/%m/%Y%m_{key}',
                        LEVEL_DAY: '%Y/%m/%d/%Y%m%d_{key}',
                        LEVEL_HOUR: '%Y/%m/%d/%H/%Y%m%d%H_{key}'}

    def __init__(self, root, level=None, pattern=None, mode='a', open_files_timeout=600, buffer_size=10000, buffer_timeout=60):
        self.root = self._fix_root(root)

        self.mode = mode
        if mode == 'w':
            remove_path(self.root)
        make_directories(self.root)

        # put level and pattern into metadata
        # there is 2 formats for granular name due to refactoring bug
        fn1 = os.path.join(self.root, '.gran')
        fn2 = os.path.join(self.root, '.gran_df')
        self.metadata_fn = fn2 if os.path.exists(fn2) else fn1

        self.metadata = {}
        if os.path.exists(self.metadata_fn):
            self.load_metadata()

            if level is not None and self.level != level:
                raise Exception(f'Level is already initialized and is different {self.root}')

            if pattern and self.pattern != pattern:
                raise Exception(f'Pattern is already initialized and is different {self.root}')
        else:
            if level is None:
                raise Exception('Granular not initialized. Specify level')
            self.level = level
            self.pattern = pattern or self.default_patterns[level]
            self.pattern += self.ext
            self.dump_metadata()

        self.level_keys = LEVEL_KEYS[:self.level]

        self.open_files_timeout = cast_timedelta(open_files_timeout)

        if not self._is_pattern_valid(self.pattern):
            raise Exception(f'Pattern {self.pattern} is not valid for level {self.level}')

        self._struct_df = None

        # stream
        # {fn: `HDFFile`}
        self.files = {}
        # {fn: `datetime`}
        self.last_used = {}
        self.buffer_size = buffer_size
        self.buffer_timeout = buffer_timeout

    @property
    def ext(self):
        raise NotImplementedError('Extension not specified')

    def load_metadata(self):
        with open(self.metadata_fn, 'r') as f:
            self.metadata = json.load(f)
        self.level = self.metadata['level']
        self.pattern = self.metadata['pattern']

    def dump_metadata(self):
        self.metadata.update({'level': self.level,
                              'pattern': self.pattern,
                              'ext': self.ext,
                              'storage_type': 'granular'})
        with open(self.metadata_fn, 'w') as f:
            json.dump(self.metadata, f, indent=1)

    def upd_metadata(self, md):
        self.metadata.update(md)
        self.dump_metadata()

    def init_file(self, fn, buffer_size=10000, buffer_timeout=60):
        raise NotImplementedError('Not implemented')

    def iter_granularity(self, df):
        # todo: optimize: do not create a copy!
        df = df.copy_path()

        granular_keys = [f'_{i}' for i in range(self.level)]
        full_keys = ['year', 'month', 'day', 'hour']
        for i in range(self.level):
            df[granular_keys[i]] = getattr(df.index, full_keys[i])

        if self.level == LEVEL_ZERO:
            fn = os.path.join(self.root, self.pattern)
            yield fn, df
        else:
            for ind, grp in df.groupby(granular_keys):
                ind = ind if isinstance(ind, (list, tuple)) else [ind]
                # fit date
                fn = self.pattern
                for i in range(self.level):
                    fn = fn.replace(self.level_keys[i], str(ind[i]).zfill(2))
                fn = os.path.join(self.root, fn)
                # todo: optimize drop - no copying!
                yield fn, grp.drop(granular_keys, axis=1)

    def remove(self, key):
        struct_df = self.get_struct_df(key)
        fns = struct_df['fn'].unique()
        if fns:
            for fn in fns:
                os.remove(fn)
            self.reset_struct_df()

    def append(self, df, key, append=True, append_single=True):
        """
        :param df: `pd.DataFrame`
        :param key: str
        :param table: str
        :param append: bool. When False, the function will delete table, before appending new granular_storage
        :param append_single: bool. Similar to append, but is relevant only to the files, that are relevant for current df (withing it's timerange)
        :return:
        """
        if not append:
            self.remove(key)

        for fn, grp in self.iter_granularity(df):
            fn = fn.format(key=key)
            self._makedirs(fn)
            self._append_single(grp, fn, append_single)

        # todo: optimize
        self.reset_struct_df()

    def _append_single(self, df, fn, append=True):
        with self.init_file(fn) as f:
            f.append(df, append=append)

    # stream
    def append_stream(self, df, key):
        for fn, grp in self.iter_granularity(df):
            fn = fn.format(key=key)
            self._makedirs(fn)
            if fn not in self.files:
                self.files[fn] = self.init_file(fn, buffer_size=self.buffer_size, buffer_timeout=self.buffer_timeout)
            self.files[fn].append_stream(grp)
            self.last_used[fn] = datetime.now()
        self.close_stalled_files()

        # todo: optimize
        self.reset_struct_df()

    def close_stalled_files(self):
        for fn, f in dict(self.files).items():
            if fn in self.last_used and datetime.now() - self.last_used[fn] > self.open_files_timeout:
                f.close()
                self.files.pop(fn)
                self.last_used.pop(fn, None)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        for f in self.files.values():
            f.close()
        self.files = {}
        self.last_used = {}

    def flush(self):
        for f in self.files.values():
            f.flush_buffer()
        self.reset_struct_df()

    def __getitem__(self, key):
        return self.read(key)

    def read(self, key, beg=None, end=None):
        gran_df = self.get_struct_df(key)
        if self.level == LEVEL_ZERO:
            gran_df = gran_df[gran_df['ts'] == '']

            if gran_df is None or len(gran_df) == 0:
                raise NoFilesException('No files found at location for specified period')

            if len(gran_df) != 1:
                raise Exception('Bad file structure')

            fn = gran_df.iloc[0]['fn']
            with self.init_file(fn) as f:
                return f.read()

        if gran_df is None or len(gran_df) == 0:
            raise NoFilesException('No files found at location for specified period')

        # todo: make properly
        ts_len = len(gran_df['ts'].iloc[0])

        # filter tables by beg and end
        if beg:
            try:
                beg = cast_datetime(beg).strftime('%Y%m%d%H%M%S')
            except:
                if not isinstance(beg, str):
                    # beg may be like 201801, not a datetime object
                    raise
            beg = beg[:ts_len]
            gran_df = gran_df[gran_df['ts'] >= beg]

        if end:
            try:
                end = cast_datetime(end).strftime('%Y%m%d%H%M%S')
            except:
                if not isinstance(end, str):
                    # end may be like 201801, not a datetime object
                    raise
            end = end[:ts_len]
            gran_df = gran_df[gran_df['ts'] <= end]

        if len(gran_df) == 0:
            raise NoFilesException('No files found at location for specified period')

        # collect and concat all dataframes
        dfs = []
        for fn, grp in gran_df.groupby(by='fn'):
            with self.init_file(fn) as f:
                dfs.append(f.read())
        df = pd.concat(dfs, axis=0)
        return df

    def merge(self, df, key, by=None, by_index=False, keep='last'):
        """
        :param df:
        :param key:
        :param by:
        :param keep: str. 'last' or 'first' (keep old or new duplicated values)
        :return:
        """
        beg, end = df.index[0], df.index[-1]
        try:
            # beg and end are inclusive here
            cur_df = self.read(key, beg, end)
        except NoFilesException:
            cur_df = pd.DataFrame()

        if len(cur_df) > 0:
            df = merge([cur_df, df], by=by, by_index=by_index, keep=keep)

        self.append(df, key, append_single=False)

    def reset_struct_df(self):
        self._struct_df = None

    def get_struct_df(self, key=None):
        if self._struct_df is None:
            self.upd_struct()
        df = self._struct_df
        if key:
            df = df[df['key'] == key]
        return df.sort_values(by='ts')

    def keys(self):
        return list(self.get_struct_df()['key'].unique())

    def upd_struct(self):
        """ Update self._struct_df for reading. """
        values = []
        for fn in list_files(self.root, recursive=True):
            if self._is_system_file(fn.replace(self.root, '')):
                continue

            fn = self._fix_sep(fn)

            res = self._parse_granularity(fn)
            if not res:
                continue
            values.append(res)

        df = pd.DataFrame(values, columns=['fn', 'ts', 'key'])
        self._struct_df = df
        return self._struct_df

    @staticmethod
    def _is_system_file(fn):
        while fn:
            if os.path.basename(fn) in ['.gran_df', '.stream']:
                return True
            fn = os.path.dirname(fn)
        return False

    def _parse_granularity(self, path):
        """
        :param path: '20180315_filename.h5/some/table' -> [<full_path>, '20180315', filename.h5, /some/table] if self.pattern = '%Y%m%d_{fn}/{table}'
        :return:
        """
        y_pat = '\d{4}'
        m_pat = '\d{2}'
        d_pat = '\d{2}'
        h_pat = '\d{2}'
        key_pat = '[^/]+'
        ext_pat = f'\.{self.ext[1:]}$'

        pat = self.pattern
        pat = pat[:-len(self.ext)]
        pat = re_tools.replace_with_pattern(pat, '%Y', 'Y', y_pat)
        pat = re_tools.replace_with_pattern(pat, '%m', 'm', m_pat)
        pat = re_tools.replace_with_pattern(pat, '%d', 'd', d_pat)
        pat = re_tools.replace_with_pattern(pat, '%H', 'H', h_pat)
        pat = pat.replace('{key}', re_tools.def_group('key', key_pat))
        pat = pat + ext_pat

        base_path = path.replace(self.root, '')
        search = re.search(pat, base_path)
        if not search:
            return
        group = search.groupdict()

        ts = ''.join([group[level_key.replace('%', '')] for level_key in self.level_keys])
        return [path, ts, group['key']]

    def _fix_sep(self, path):
        path = path.replace('\\', '/')
        path = path.replace(os.path.sep, '/')
        return path

    def _fix_root(self, root):
        if root[-1] != '/':
            root += '/'
        return self._fix_sep(root)

    def _makedirs(self, path):
        cur_dir = os.path.dirname(path)
        if not os.path.exists(cur_dir):
            os.makedirs(cur_dir)

    def _is_pattern_valid(self, pattern):
        # todo: more extensive check
        if not all([level_key in pattern for level_key in self.level_keys]):
            return False
        if '{key}' not in pattern:
            return False
        return True


class GranularCSV(Granular):
    def init_file(self, fn, buffer_size=10000, buffer_timeout=60):
        return CSVFile(fn, buffer_size, buffer_timeout, compression=None)

    @property
    def ext(self):
        return '.csv'


class GranularFeather(Granular):
    def init_file(self, fn, buffer_size=1000, buffer_timeout=60):
        return FeatherFile(fn, buffer_size, buffer_timeout)

    @property
    def ext(self):
        return '.feather'


class GranularMsgPack(Granular):
    def init_file(self, fn, buffer_size=1000, buffer_timeout=60):
        return MsgPackFile(fn, buffer_size, buffer_timeout)

    @property
    def ext(self):
        return '.msg'


class GranularParquet(Granular):
    def init_file(self, fn, buffer_size=1000, buffer_timeout=60):
        return ParquetFile(fn, buffer_size, buffer_timeout)

    @property
    def ext(self):
        return '.parquet'


def ex_basic(granular=GranularCSV, key1='granular_storage', key2='data2'):
    import numpy as np
    print('Basic usage test')

    # note: mode='w' will delete all existing .h5 files in root directory! Be careful.
    gran = granular('granular_storage/root/', LEVEL_DAY, mode='w')
    index = [cast_datetime('2018.05.01') + timedelta(days=1) * i for i in range(3)]
    df1 = pd.DataFrame(np.random.randn(3, 1), index=index, columns=['0'])
    df2 = pd.DataFrame(np.random.randn(3, 2), index=index, columns=['0', '1'])
    print('Initial dataframes')
    print(df1)
    print(df2)

    print('Read')
    gran.append(df1, key1)

    print(gran[key1])
    gran.append(df2, key1)

    print(gran.read(key1))
    # or with __getitem__. They are the same
    print(gran[key1])

    print('Read with beg and end')
    print(gran.read(key1, beg='2018.05.01', end='2018.05.02'))

    print('Get keys')
    print(gran.keys())

    print('Write and read with extended granular_storage')
    gran.append(df2, key1)

    print(gran[key1])

    print('Level zero test')
    gran = granular('granular_storage/root/', LEVEL_ZERO, mode='w')
    gran.append(df1, key1)
    print(gran.get_struct_df(key1))
    print(gran[key1])
    gran.append(df2, key1)
    print(gran[key1])

    print('Stream examples')
    with granular('granular_storage/root/', LEVEL_DAY, mode='w') as gran:
        gran.append_stream(df1, key1)
        gran.append_stream(df1, key1)
        gran.append_stream(df2, key1)
        gran.append_stream(df2, key1)

    # with statement handles flush properly
    gran = granular('granular_storage/root/', LEVEL_DAY)
    print(gran[key1])

    print('Custom pattern')
    gran = granular('granular_storage/root', LEVEL_DAY, pattern='%Y/%Y%m%d_{key}', mode='w')
    gran.append(df1, key1)
    print(gran[key1])
    print(gran.get_struct_df())
    print(gran.get_struct_df().groupby('key').agg('last'))

    # append - false
    gran = granular('granular_storage/root', LEVEL_DAY, mode='w')
    tmp = df1.iloc[1:].copy_path()

    # test append=False
    print("Remove")
    gran.append(df1, key1)
    gran.append(tmp, key1, append=False)
    print(gran[key1])
    print(gran.get_struct_df())

    gran.remove(key1)
    print(gran.get_struct_df())

    print("Remove 2")
    gran.append(df1, key1)
    gran.append(df1, key2)
    gran.append(tmp, key1, append=False)
    print(gran[key1])
    print(gran[key2])
    print(gran.get_struct_df())

    print('Merge')
    gran = granular('granular_storage/root', LEVEL_DAY, mode='w')
    gran.append(df1, key1)
    tmp = df1.copy()
    tmp[:] = 0
    tmp = tmp.iloc[1:]
    gran.merge(tmp, key1, by_index=True, keep='last')
    print(gran[key1])


if __name__ == '__main__':
    print('GranularCSV' + ' - ' * 100)
    print()
    ex_basic(GranularCSV, 'granular_storage', 'data2')
    print()
    print('GranularFeather' + ' - ' * 100)
    print()
    ex_basic(GranularFeather, 'granular_storage', 'data2')
    print()
    print('GranularMsgPack' + ' - ' * 100)
    print()
    ex_basic(GranularMsgPack, 'granular_storage', 'data2')



import pandas as pd
from datetime import datetime
from utils_ak.time.dt import cast_timedelta


# todo: specify columns and data_columns for efficiency
# todo: inherit from streamfile somehow


class HDFFile(pd.HDFStore):
    """ A simple wrapper for pd.HDFStore for dumping streaming granular_storage. """

    def __init__(self, path, mode='a', complevel=9, complib='blosc:lz4hc', buffer_size=None, buffer_timeout=60, extensible=True, **kwargs):
        self.buffer_size = buffer_size or 0
        self.cur_size = 0
        self.buffer_timeout = cast_timedelta(buffer_timeout)

        # {key: [value1, value2, ...]}
        self.buffer = {}
        self.last_flush = datetime.now()
        self.extensible = extensible
        super().__init__(path, mode, complevel, complib, **kwargs)

    def close(self):
        self.flush_buffer()
        super().close()

    def append_stream(self, df, table):
        if len(df) == 0:
            raise Exception('Cannot add an empty dataframe')

        self.buffer.setdefault(table, []).append(df)
        self.cur_size += len(df)

        if self.is_buffer_full():
            self.flush_buffer()

    def is_buffer_full(self):
        if self.buffer_timeout:
            if datetime.now() - self.last_flush > self.buffer_timeout:
                return True
        return self.cur_size > self.buffer_size

    def append(self, df, table, append=True, **kwargs):
        try:
            super().append(table, df, format='table', append=append, **kwargs)
        except ValueError:
            if not self.extensible:
                raise
            df = pd.concat([self[table], df], axis=0)
            super().append(table, df, format='table', append=False, **kwargs)

    def flush_buffer(self):
        for table, dfs in self.buffer.items():
            df = pd.concat(dfs, axis=0)
            self.append(df, table)

        self.buffer = {}
        self.last_flush = datetime.now()
        self.cur_size = 0

    def getdefault(self, key, default=None):
        if key in self:
            return self[key]
        return default


def ex():
    import numpy as np
    from datetime import datetime, timedelta
    import os

    if not os.path.exists('granular_storage/'):
        os.makedirs('granular_storage/')

    df1 = pd.DataFrame(np.random.randn(2, 1), index=[datetime.now() + timedelta(days=1) * i for i in range(2)])
    print('Dataframe 1')
    print(df1)

    with HDFFile('granular_storage/test.h5', mode='w', buffer_size=3) as hdf:
        hdf.append_stream(df1, 'backtest/trades')
        hdf.append_stream(df1, 'backtest/trades')
        # flushing is here
        hdf.append_stream(df1, 'backtest/trades')
        print('Current keys: {}'.format(hdf.keys()))

    # flushing is on the close of file

    with HDFFile('granular_storage/test.h5') as hdf:
        # 6 values here
        print(hdf['backtest/trades'])
        print(hdf.keys())

    # extension test
    df2 = pd.DataFrame(np.random.randn(2, 2), index=[datetime.now() + timedelta(days=1) * i for i in range(2)])

    with HDFFile('granular_storage/test.h5', mode='a') as hdf:
        hdf.append_stream(df2, 'backtest/trades')
        print(hdf['backtest/trades'])

    with HDFFile('granular_storage/test.h5', mode='a') as hdf:
        hdf.append_stream(df2, 'backtest/trades')
        print(hdf['backtest/trades'])


if __name__ == '__main__':
    ex()



import os
import openpyxl as opx
from openpyxl.styles import Alignment, PatternFill, Font
from openpyxl.styles.borders import Border, Side, BORDER_THIN
from openpyxl.utils import get_column_letter
from utils_ak.color import cast_color


def init_workbook(sheet_names=None, active_sheet_name=None):
    workbook = opx.Workbook()
    sheet_names = sheet_names or []
    for i, sheet_name in enumerate(sheet_names):
        workbook.create_sheet(sheet_name, i)
    workbook.remove(workbook.worksheets[-1])
    if active_sheet_name:
        workbook.active = sheet_names.index(active_sheet_name)
    return workbook


def cast_workbook(wb_obj):
    if isinstance(wb_obj, str):
        return opx.load_workbook(filename=wb_obj, data_only=True)
    elif isinstance(wb_obj, opx.Workbook):
        return wb_obj
    elif isinstance(wb_obj, list):
        return init_workbook(sheet_names=wb_obj)
    else:
        raise Exception('Unknown workbook format')

def set_border(sheet, x, y, w, h, border):
    rows = sheet['{}{}'.format(get_column_letter(x), y):'{}{}'.format(get_column_letter(x + w - 1), y + h - 1)]

    for row in rows:
        row[0].border = Border(left=border, top=row[0].border.top, bottom=row[0].border.bottom, right=row[0].border.right)
        row[-1].border = Border(left=row[-1].border.left, top=row[-1].border.top, bottom=row[-1].border.bottom, right=border)
    for c in rows[0]:
        c.border = Border(left=c.border.left, top=border, bottom=c.border.bottom, right=c.border.right)
    for c in rows[-1]:
        c.border = Border(left=c.border.left, top=c.border.top, bottom=border, right=c.border.right)


def draw_cell(sheet, x, y, text, color=None, font_size=None, text_rotation=None, alignment=None):
    cell = sheet.cell(row=y, column=x)
    cell.font = Font(size=font_size)
    if alignment == 'center':
        cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True, text_rotation=text_rotation)
    cell.value = text

    if color:
        cell.fill = PatternFill("solid", fgColor=color[1:])
    return cell


def draw_block(sheet, x1, x2, h1, h2, text, color=None, border=None, text_rotation=None, font_size=None, alignment=None, bold=False):
    color = color or cast_color('white')
    sheet.merge_cells(start_row=x2, start_column=x1, end_row=x2 + h2 - 1, end_column=x1 + h1 - 1)
    merged_cell = sheet.cell(row=x2, column=x1)
    merged_cell.font = Font(size=font_size, bold=bold)
    if alignment == 'center':
        merged_cell.alignment = Alignment(horizontal='center', vertical='center', wrap_text=True, text_rotation=text_rotation)
    merged_cell.value = text
    if color:
        merged_cell.fill = PatternFill("solid", fgColor=color[1:])

    if border is not None:
        if isinstance(border, dict):
            border = Side(**border)
        elif isinstance(border, Side):
            pass
        else:
            raise Exception('Unknown border type')

        set_border(sheet, x1, x2, h1, h2, border)


def draw_row(sheet, y, values, color=None, **kwargs):
    for i, v in enumerate(values, 1):
        draw_cell(sheet, i, y, text=v, color=color, **kwargs)


if __name__ == '__main__':
    wb = init_workbook(['a', 'b'], active_sheet_name='b')
    wb.save('text.xlsx')



""" Granular is a reader/writer for granular granular_storage format. """
from .granular import GranularMsgPack, GranularFeather, GranularCSV, NoFilesException, GranularParquet
from .storage import GranularStorage, GranularCSVStorage, GranularFeatherStorage, GranularMsgPackStorage, GranularParquetStorage
from .sync import GranularSync
from .streamer import GranularStreamer
from .enums import *



from utils_ak.simple_microservice.microservice import SimpleMicroservice
import multiprocessing


class Listener(SimpleMicroservice):
    def __init__(self, collection, topic='', *args, **kwargs):
        super().__init__(f'Listener_{collection}', *args, **kwargs)
        self.add_callback(collection, topic, self._log, formatter='default')

    def _log(self, topic, msg):
        self.logger.info(f'{topic}-{msg}')


def run_listener(collection, topic='', *args, **kwargs):
    Listener(collection, topic, *args, **kwargs).run()


def run_listener_async(collection, topic='', *args, **kwargs):
    args = [collection, topic] + list(args)
    multiprocessing.Process(target=run_listener, args=args, kwargs=kwargs).start()



import sys
from loguru import logger


def configure_loguru_stdout(level='DEBUG'):
    logger.add(sys.stdout, level=level, format='<green>{time:YYYY-MM-DD HH:mm:ss!UTC}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan> - <level>{message}</level> - <yellow>{extra}</yellow>')



import copy
from datetime import datetime

from bson.objectid import ObjectId
from mongoengine import *

from utils_ak.time import cast_str
from utils_ak.mongoengine import *

from datetime import datetime


class Job(Document):
    AUTO_FIELDS = ['created', '_id']
    type = StringField(required=True)
    payload = DictField()

    created = DateTimeField(default=datetime.utcnow)

    locked_by = ReferenceField('Execution')
    locked_at = DateTimeField()

    workers = ListField(ReferenceField('Worker'))

    meta = {'allow_inheritance': True}


class Worker(Document):
    AUTO_FIELDS = ['created', '_id']
    job = ReferenceField(Job)
    config = DictField()
    created = DateTimeField(default=datetime.utcnow)
    status = StringField(required=True, default='pending', choices=['pending', 'initializing', 'running', 'error', 'success', 'stalled'])
    response = StringField()

    meta = {'allow_inheritance': True}



from utils_ak.granular_storage.table.timeseries.fs import FsTimeSeriesTable
from utils_ak.os_tools import makedirs, open_atomic, remove


class BasicFsTimeSeriesTable(FsTimeSeriesTable):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn
        makedirs(self.fn)

        self._f = None

    def encode(self, msg):
        raise NotImplementedError

    def read_file(self, f):
        raise NotImplementedError

    @property
    def f(self):
        if not self._f:
            self._f = open(self.fn, 'ab')
        return self._f

    def store(self, ts, key, msg):
        self.f.write(self.encode(msg))

    def flush(self):
        self.f.flush()

    def close(self):
        if self._f:
            self._f.close()
            self._f = None

    def clear(self):
        self.close()
        remove(self.fn)

    def store_many(self, values, safe=False, overwrite=False):
        """
        :param values: [[ts, key, msg], ...]
        :param safe: will make copy first, write to copy and then rename to initial file. Useful when we don't want granular_storage to be crashed at any cost.
        :return:
        """
        # remove ts and key from values (not used)
        values = [value[2] for value in values]

        if overwrite:
            self.clear()

        if not safe:
            for msg in values:
                self.store(None, None, msg)
        else:
            # append old table first
            self.close()
            values = list(self.read()) + values

            with open_atomic(self.fn, 'wb', fsync=True) as f:
                for msg in values:
                    f.write(self.encode(msg))

    def read(self):
        raise NotImplementedError



from utils_ak.fluid_flow import *


def test_flow_queue_1_in():
    parent = Container('Parent', value=100, max_pressures=[None, 20])

    child1 = Container('Child1', max_pressures=[20, None], limits=[40, None])
    child2 = Container('Child2', max_pressures=[10, None], limits=[50, None])

    queue = Queue('Queue', [child1, child2])

    pipe_connect(parent, queue, 'parent-queue')

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_queue_2_out():
    parent1 = Container('Parent1', value=100, max_pressures=[None, 10], limits=[None, 100])
    parent2 = Container('Parent2', value=100, max_pressures=[None, 20], limits=[None, 100])
    queue = Queue('Parent', [parent1, parent2])

    child = Container('Child', max_pressures=[None, None])
    pipe_connect(queue, child, 'parent-queue')

    flow = FluidFlow(queue, verbose=True)
    run_flow(flow)


def test_flow_queue_3():
    parent = Container('Parent', value=100, max_pressures=[None, 20])

    child1 = Processor('Child1', max_pressures=[20, None], processing_time=5, limits=[40, None])
    child2 = Processor('Child2', max_pressures=[10, None], processing_time=5, limits=[50, None])

    queue = Queue('Queue', [child1, child2])

    pipe_connect(parent, queue, 'parent-queue')

    flow = FluidFlow(parent, verbose=True)
    run_flow(flow)


def test_flow_queue_4_different_items():
    parent1 = Container('Parent1', item='a', value=100, max_pressures=[None, 10], limits=[None, 100])
    parent2 = Container('Parent2', item='b', value=100, max_pressures=[None, 20], limits=[None, 100])
    queue = Queue('Parent', [parent1, parent2])

    hub = Hub('hub')
    child1 = Container('Child1', item='a', max_pressures=[None, None])
    child2 = Container('Child2', item='b', max_pressures=[None, None])

    pipe_connect(queue, hub, 'parent-hub')
    pipe_connect(hub, child1, 'hub-child1')
    pipe_connect(hub, child2, 'hub-child2')
    flow = FluidFlow(queue, verbose=True)
    run_flow(flow)


if __name__ == '__main__':
    test_flow_queue_1_in()
    # test_flow_queue_2_out()
    # test_flow_queue_3()
    # test_flow_queue_4()



from datetime import datetime
from pprint import pprint


class MonitorActor:
    def __init__(self, microservice, heartbeat_timeout=10):
        self.microservice = microservice
        self.workers = {}  # {id: {status, state, last_heartbeat}}
        self.heartbeat_timeout = heartbeat_timeout

        self.microservice.add_callback('monitor', '', self.on_monitor)
        self.microservice.add_timer(self.update_stalled, 3.0)

        # todo: del, hardcode
        self.received_messages_cache = []

    def update_status(self, worker_id, status):
        if status != self.workers[worker_id].get('status'):
            self.microservice.publish_json('monitor_out', 'status_change', {'id': worker_id, 'old_status': self.workers[worker_id].get('status'), 'new_status': status})
            self.workers[worker_id]['status'] = status

    def update_stalled(self):
        for worker_id in self.workers:
            if 'status' not in self.workers[worker_id]:
                # non-initialized
                # todo: make properly
                continue

            if self.workers[worker_id]['status'] == 'success':
                continue

            if 'last_heartbeat' in self.workers[worker_id] and (datetime.utcnow() - self.workers[worker_id]['last_heartbeat']).total_seconds() > self.heartbeat_timeout:
                self.update_status(worker_id, 'stalled')

    def on_monitor(self, topic, msg):
        # todo: preview hardcode
        if 'state' in msg:
            key = msg['id'] + str(msg['state'])
            if key not in self.received_messages_cache:
                self.received_messages_cache.append(key)
            else:
                return

        self.microservice.logger.info('On monitor', topic=topic, msg=msg)

        worker_id = msg['id']
        if worker_id not in self.workers:
            self.microservice.publish_json('monitor_out', 'new', {'id': worker_id})
            self.workers[worker_id] = {}

        if topic == 'heartbeat':
            self.workers[worker_id]['last_heartbeat'] = datetime.utcnow()
        elif topic == 'state':
            self.update_status(worker_id, msg['status'])
            self.workers[worker_id]['state'] = msg['state']
        elif topic in ['status_change', 'new']:
            pass
        else:
            raise ValueError(topic)

        self.microservice.logger.info('Current monitor state', state=self.workers)


""" Email functionality. """
import pyzmail
import json
from smtplib import SMTP_SSL
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
import email.encoders as Encoders


def send_gmail(username, password, to_, subject, text, html=False):
    # compose mail
    sender = (u'I3K', username)
    # recipients = [(u'Him', to_), username]
    if type(to_) == str:
        to_ = [to_]
    if type(to_) != list:
        raise Exception('wrong to_ format')

    recipients = [(u'Good Guy', t) for t in to_]
    subject = subject
    text_content = text
    # prefered_encoding = 'iso-8859-1'
    # text_encoding = 'iso-8859-1'
    prefered_encoding = 'utf-8'
    text_encoding = 'utf-8'

    if html:
        payload, mail_from, rcpt_to, msg_id = pyzmail.compose_mail(sender, recipients, subject, prefered_encoding, None, html=(text, text_encoding), attachments=None)
    else:
        payload, mail_from, rcpt_to, msg_id = pyzmail.compose_mail(sender, recipients, subject, prefered_encoding, (text_content, text_encoding), html=None, attachments=None)

    # send mail
    smtp_host = 'smtp.gmail.com'
    smtp_port = 587
    smtp_mode = 'tls'
    smtp_login = username
    smtp_password = password

    # sends to our mail too
    # pyzmail.send_mail(payload, mail_from, rcpt_to, smtp_host, smtp_port=smtp_port, smtp_mode=smtp_mode, smtp_login=smtp_login, smtp_password=smtp_password)

    return pyzmail.send_mail(payload, None, rcpt_to, smtp_host, smtp_port=smtp_port, smtp_mode=smtp_mode, smtp_login=smtp_login, smtp_password=smtp_password)


def send_yandex(username, password, to_, subject, text, html=False, headers=[], sender_name=u'Good boy'):
    # compose mail
    sender = (sender_name, username)
    # recipients = [(u'Him', to_), username]
    if type(to_) == str:
        to_ = [to_]
    if type(to_) != list:
        raise Exception('wrong to_ format')

    recipients = [(t, t) for t in to_]
    subject = subject
    text_content = text
    # prefered_encoding = 'iso-8859-1'
    # text_encoding = 'iso-8859-1'
    prefered_encoding = 'utf-8'
    text_encoding = 'utf-8'

    if html:
        payload, mail_from, rcpt_to, msg_id = pyzmail.compose_mail(sender, recipients, subject, prefered_encoding, None, html=(text, text_encoding), attachments=None, headers=headers)
    else:
        payload, mail_from, rcpt_to, msg_id = pyzmail.compose_mail(sender, recipients, subject, prefered_encoding, (text_content, text_encoding), html=None, attachments=None, headers=headers)

    # send mail
    smtp_host = 'smtp.yandex.ru'
    smtp_port = 0
    smtp_mode = 'ssl'
    smtp_login = username
    smtp_password = password

    return pyzmail.send_mail(payload, username, rcpt_to, smtp_host, smtp_port=smtp_port, smtp_mode=smtp_mode, smtp_login=smtp_login, smtp_password=smtp_password)


if __name__ == '__main__':
    msg = dict(a=u'аладушек', b=2)
    # print(send_gmail('email_from', 'password_here', ['emails_to'], 'subject', json.dumps(msg, indent=2, ensure_ascii=False).encode('utf-8')))
    print(send_yandex('rainydai@yandex.ru', 'rlicu12MLP&$*', ['arseniikadaner@gmail.com'], 'subject', json.dumps(msg, indent=2, ensure_ascii=False).encode('utf-8')))



import os

from utils_ak.os.os_tools import make_directories
from utils_ak.time.dt import cast_datetime
from utils_ak.dict.limited_dict import LimitedDict

import time


class Storage:
    def __init__(self, root):
        self.root = root
        make_directories(root)

    def __getitem__(self, db_name):
        return self.get_db(db_name)

    @property
    def table_cls(self):
        raise NotImplemented

    @property
    def db_cls(self):
        raise NotImplemented

    @property
    def ext(self):
        return '.txt'

    def get_db(self, db_name, *args, **kwargs):
        return self.db_cls(os.path.join(self.root, db_name), self.table_cls, self.ext, *args, **kwargs)


class DB:
    def __init__(self, path, table_cls, ext):
        self.path = path
        self.table_cls = table_cls
        self.ext = ext

    def get_table(self, table_name, *args, **kwargs):
        return self.table_cls(os.path.join(self.path, table_name + self.ext), *args, **kwargs)

    def __getitem__(self, table_name):
        return self.get_table(table_name)


class GranularTable(object):
    """ This class is a decorator that makes any table granulared """
    def __init__(self, fn, pattern, cache_size=100, *args, **kwargs):
        self.table_args = args
        self.table_kwargs = kwargs

        self.dirname = os.path.dirname(fn)
        self.basename = os.path.basename(fn)
        self.base, self.ext = os.path.splitext(self.basename)
        self.pattern = pattern
        self.tables = {}

        self._memo = LimitedDict(cache_size)

    def store(self, ts, key, message):
        if ts is None:
            ts = time.time()
        self._get_atom_table(ts, key).store(ts, key, message)

    def store_many(self, values, *args, **kwargs):
        values_by_table = {}
        for ts, key, msg in values:
            table = self._get_atom_table(ts, key)
            values_by_table.setdefault(table, []).append([ts, key, msg])

        for table, table_values in values_by_table.items():
            table.store_many(table_values, *args, **kwargs)

    def _get_granular_fn(self, ts, key):
        _cache_name = str(ts) + "|" + str(key)
        if _cache_name in self._memo:
            return self._memo[_cache_name]

        fn = cast_datetime(ts).strftime(self.pattern)
        fn = fn.format(key=key, base=self.base)
        result = os.path.join(self.dirname, fn + self.ext)
        self._memo[_cache_name] = result
        return result

    def _get_atom_table(self, ts, key):
        fn = self._get_granular_fn(ts, key)
        if fn not in self.tables:
            self.tables[fn] = self.table_cls(fn, *self.table_args, **self.table_kwargs)
        return self.tables[fn]

    def close(self):
        for name, table in self.tables.items():
            table.close()

    def flush(self):
        for name, table in self.tables.items():
            table.flush()

    @property
    def table_cls(self):
        raise NotImplemented()





import multiprocessing

from utils_ak.simple_microservice import *


class WorkerMicroservice(SimpleMicroservice):
    def __init__(self, id, *args, **kwargs):
        super().__init__(id, *args, **kwargs)
        self.add_timer(self.publish_json, 3.0, args=('monitor', 'heartbeat', {'id': self.id},))

    def send_state(self, status, state):
        self.publish_json('monitor', 'state', {'id': self.id, 'status': status, 'state': state})


def test():
    run_listener_async('monitor', message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))
    ms = WorkerMicroservice('Worker', message_broker=('zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}}}))
    ms.run()


if __name__ == '__main__':
    test()


""" Miscellaneous functionality with basic python built-in objects. """
import collections
import numpy


def delistify_single_list(lst_obj):
    if isinstance(lst_obj, list):
        return lst_obj if (len(lst_obj) > 1 or not lst_obj) else lst_obj[0]
    else:
        return lst_obj


delistify = delistify_single_list


def listify(obj):
    if not isinstance(obj, list):
        return [obj]
    return obj


def list_get(lst, index, default=None):
    try:
        return lst[index]
    except IndexError:
        return default


def cast_list(lst):
    return lst if isinstance(lst, list) else [lst]


def remove_duplicates(seq, key=None):
    # preserves order of sequence
    seen = set()
    seen_add = seen.add
    key = key or (lambda x: x)
    return [x for x in seq if not (key(x) in seen or seen_add(key(x)))]


def remove_neighbor_duplicates(seq, key=None):
    key = key or (lambda x: x)
    res = []
    for x in seq:
        if res and key(res[-1]) == key(x):
            continue
        res.append(x)
    return res


def apply_on_slice(f, lst, cond):
    ind = [i for i, v in enumerate(lst) if cond(v)]
    applied = f([lst[i] for i in ind])
    res = list(lst)
    for j, i in enumerate(ind):
        res[i] = applied[j]
    return res

# NOTE: use anyconfig.merge for more advanced merge logics
def update_dic(dic, new_dic):
    """
    :param dic: obj, dict by default
    :param new_dic: dict
    :return:
    """
    for key, val in new_dic.items():
        if isinstance(dic, collections.Mapping):
            if isinstance(val, collections.Mapping):
                dic[key] = update_dic(dic.get(key, {}), val)
            else:
                dic[key] = new_dic[key]
        else:
            # dic is not a dictionary. Make it one
            dic = {key: new_dic[key]}
    return dic


def chunks(lst, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def split_list(lst, n):
    """Split list to n parts with equal size."""
    for array in numpy.array_split(lst, n):
        yield list(array)


def unfold_dic(dic, keys, get=False, default=None):
    if get:
        return [dic.get(key, default) for key in keys]
    else:
        return [dic[key] for key in keys]

def test_collection():
    d1 = {'k1': {'k3': 4}}
    d2 = {'k1': {'k3': [2, 3]}}
    print(update_dic(d1, d2), d1, d2)

    print(apply_on_slice(lambda lst: [x ** 2 for x in lst], [1,2,3,4,5,6], lambda x: x % 2 == 0))
    # print(filter_dic({'k': 1, 'e': 2}, leave=2))
    print(remove_duplicates([1,2,3, 4,5,1,1,]))
    print(remove_neighbor_duplicates([1,2,3, 4,5,1,1,]))

if __name__ == '__main__':
    test_collection()


from functools import partial
from utils_ak.simple_vector import *
from utils_ak.properties import *
from utils_ak.block_tree import Block

from utils_ak.clock import *


def x_cumsum_acc(parent, child, key, default=None, formatter=None):
    if parent:
        return parent[key].add(relative_acc(parent, child, key, default=default, formatter=formatter))
    return SimpleVector(list(relative_acc(parent, child, key, default=default, formatter=formatter).values))


class ParallelepipedBlock(Block):
    def __init__(self, block_class, n_dims=2, **props):
        self.n_dims = n_dims

        if 'x' not in props:
            props['x'] = SimpleVector(n_dims)
        if 'size' not in props:
            props['size'] = SimpleVector(n_dims)

        super().__init__(block_class, props_formatters={'x': lambda k, v: SimpleVector(v), 'size': lambda k, v: SimpleVector(v)}, **props)
        self.props.accumulators['x'] = x_cumsum_acc
        self.props.accumulators['x_rel'] = lambda parent_props, child_props, key: relative_acc(parent_props, child_props, 'x')
        self.props.accumulators['size'] = relative_acc
        self.props.accumulators['axis'] = partial(relative_acc, default=0)
        self.props.accumulators['is_parent_node'] = partial(relative_acc, default=False)

    @property
    def x(self):
        return self.props['x']

    @property
    def y(self):
        return self.x + self.size

    def __str__(self):
        res = self.props['cls'] + ' ' + ' x '.join([f'({self.x[i]}, {self.y[i]}]' for i in range(self.n_dims)])

        for child in self.children:
            for line in str(child).split('\n'):
                if not line:
                    continue
                res += '\n  ' + line
        return res

    def __repr__(self):
        return self.tabular()

    def is_leaf(self):
        return not self.children and not self.props['is_parent_node']

    def tabular(self):
        res = ''
        for b in self.iter():
            if b.size[0] != 0:
                res += ' ' * int(b.x[0]) + '=' * int(b.size[0]) + f' {b.props["cls"]} ' + ' x '.join([f'({b.x[i]}, {b.y[i]}]' for i in range(b.n_dims)])
                res += '\n'
        return res

    @property
    def size(self):
        size = self.props['size']
        values = []
        for axis in range(self.n_dims):
            if size[axis] == 0:
                if not self.children:
                    values.append(0)
                else:
                    values.append(max([c.y[axis] - self.x[axis] for c in self.children]))
            else:
                values.append(size[axis])
        return SimpleVector(values)


def test_parallelepiped_block():
    a = ParallelepipedBlock('a', n_dims=2, x=[1, 2])
    b = ParallelepipedBlock('b', n_dims=2)
    c = ParallelepipedBlock('c', n_dims=2, x=[3, 4], size=[1, 5])
    a.add_child(b)
    b.add_child(c)

    print(a)
    print(b)
    print(c)
    print(a.x, a.size, a.y)
    print(b.x, b.size, b.y)
    print(c.x, c.size, c.y)

    print()
    print(a['b']['c'])

    print(a.__repr__())


if __name__ == '__main__':
    test_parallelepiped_block()


import time
import threading

from utils_ak.mongoengine import *
from .models import Job, Worker
import logging

from utils_ak.deployment import *
from utils_ak.simple_microservice import SimpleMicroservice
from utils_ak.dict import fill_template
from utils_ak.serialization import cast_dict_or_list

from utils_ak.mongo_job_queue.monitor import MonitorActor
from utils_ak.mongo_job_queue.config import BASE_DIR


class JobOrchestrator:
    def __init__(self, deployment_controller, message_broker):
        self.timeout = 1
        self.controller = deployment_controller
        self.ms = SimpleMicroservice('JobOrchestrator', message_broker=message_broker)
        self.monitor = MonitorActor(self.ms)
        self.process_active_jobs()
        self.ms.add_timer(self.process_new_jobs, 1.0)
        self.ms.add_callback('monitor_out', '', self.on_monitor_out)

    def run(self):
        self.ms.run()

    def process_active_jobs(self):
        pass # todo: go through all active jobs and process them

    def process_new_jobs(self):
        new_jobs = Job.objects(workers__size=0).all()

        if new_jobs:
            self.ms.logger.info(f'Processing {len(new_jobs)} new jobs')

        for new_job in new_jobs:
            worker_model = Worker()
            worker_model.save()
            new_job.workers.append(worker_model)
            new_job.save()

            # generate deployment
            deployment = cast_dict_or_list(os.path.join(BASE_DIR, 'worker/deployment.yml.template'))

            # todo: Hardcode, use new_job.type
            IMAGE = 'akadaner/test-worker'
            # todo: hardcode, use generic message broker
            MESSAGE_BROKER = ['zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://host.k3d.internal:5555', 'type': 'sub'}}}]
            # MESSAGE_BROKER = ['zmq', {'endpoints': {'monitor': {'endpoint': 'tcp://host.docker.internal:5555', 'type': 'sub'}}}]

            params = {'deployment_id': str(worker_model.id), 'payload': new_job.payload, 'image': IMAGE, 'message_broker': MESSAGE_BROKER}
            deployment = fill_template(deployment, **params)
            self.ms.logger.info('Starting new worker', deployment=deployment)
            self.controller.start(deployment)

    def on_monitor_out(self, topic, msg):
        self.ms.logger.info('On monitor out', topic=str(topic), msg=str(msg))
        if topic == 'status_change':
            worker = Worker.objects(pk=msg['id']).first()  # todo: check if missing
            worker.status = msg['new_status']
            if worker.status == 'success':
                worker.response = self.monitor.workers[msg['id']]['state']['response']
            worker.save()

            if worker.status == 'success':
                self.controller.stop(msg['id'])



from utils_ak import update_dic

from bson.objectid import ObjectId


def cast_object_id(obj):
    if isinstance(obj, ObjectId):
        return obj
    elif isinstance(obj, str):
        return ObjectId(obj)
    elif obj is None:
        return None
    elif isinstance(obj, dict):
        return obj['_id']
    else:
        raise Exception('Unknown object id type')


def cast_model(obj, cls):
    if isinstance(obj, ObjectId):
        return cast_model({'_id': obj}, cls)
    elif isinstance(obj, dict):
        if '_id' in obj:
            # fetch and return updated version from server
            element = cls.objects(pk=obj['_id']).first()
            if not element:
                raise Exception('Object not found')
            db_obj = cast_model(element, cls)
            pk = obj['_id']
            d1, d2 = db_obj.to_mongo(), dict(obj)
            d1.pop('_id', None), d2.pop('_id', None)
            return cls(pk=pk, **update_dic(d1, d2))
        else:
            # init
            return cls(**obj)
    elif isinstance(obj, cls):
        return obj
    else:
        raise Exception('Unknown model format')


def cast_dict(obj, cls):
    model = cast_model(obj, cls=cls)
    res = dict(model.to_mongo())
    res['_cls'] = cls.__name__
    return res



import multiprocessing
import time
import sys

from utils_ak.deployment import *
from utils_ak.loguru import configure_loguru_stdout
from utils_ak.mongo_job_queue.job_orchestrator import JobOrchestrator
from utils_ak.mongo_job_queue.models import *

from loguru import logger

BROKER = 'zmq'
BROKER_CONFIG = {'endpoints': {'monitor': {'endpoint': 'tcp://localhost:5555', 'type': 'sub'}, 'monitor_out': {'endpoint': 'tcp://localhost:5556', 'type': 'sub'}}}
MESSAGE_BROKER = (BROKER, BROKER_CONFIG)

from mongoengine import connect


def create_new_job():
    connect(host='mongodb+srv://arseniikadaner:Nash0lsapog@cluster0.2umoy.mongodb.net/feature-store?retryWrites=true&w=majority')
    logger.info('Connected to mongodb')
    logger.remove()
    configure_loguru_stdout()
    time.sleep(2)
    logger.debug('Creating new job...')
    Job.drop_collection()
    Worker.drop_collection()
    job = Job(type='test', payload={'type': 'batch'})
    job.save()



def test():
    logger.remove()
    configure_loguru_stdout()
    connect(host='mongodb+srv://arseniikadaner:Nash0lsapog@cluster0.2umoy.mongodb.net/feature-store?retryWrites=true&w=majority')
    logger.info('Connected to mongodb')
    controller = KubernetesController()
    orchestrator = JobOrchestrator(controller, MESSAGE_BROKER)
    # multiprocessing.Process(target=create_new_job).start()
    orchestrator.run()


if __name__ == '__main__':
    test()
